"""resnet in resnet in pytorch



[1] Sasha Targ, Diogo Almeida, Kevin Lyman.

    Resnet in Resnet: Generalizing Residual Architectures
    https://arxiv.org/abs/1603.08029v1
"""

import torch
import torch.nn as nn

#geralized
class ResnetInit(nn.Module):
    def __init__(self, in_channel, out_channel, stride):
        super().__init__()

        #"""The modular unit of the generalized residual network architecture is a
        #generalized residual block consisting of parallel states for a residual stream,
        #r, which contains identity shortcut connections and is similar to the structure
        #of a residual block from the original ResNet with a single convolutional layer
        #(parameters W l,r→r )
        self.residual_stream_conv = nn.Conv2d(in_channel, out_channel, 3, padding=1, stride=stride)

        #"""and a transient stream, t, which is a standard convolutional layer
        #(W l,t→t )."""
        self.transient_stream_conv = nn.Conv2d(in_channel, out_channel, 3, padding=1, stride=stride)

        #"""Two additional sets of convolutional filters in each block (W l,r→t , W l,t→r )
        #also transfer information across streams."""
        self.residual_stream_conv_across = nn.Conv2d(in_channel, out_channel, 3, padding=1, stride=stride)

        #"""We use equal numbers of filters for the residual and transient streams of the
        #generalized residual network, but optimizing this hyperparameter could lead to
        #further potential improvements."""
        self.transient_stream_conv_across = nn.Conv2d(in_channel, out_channel, 3, padding=1, stride=stride)

        self.residual_bn_relu = nn.Sequential(
            nn.BatchNorm2d(out_channel),
            nn.ReLU(inplace=True)
        )

        self.transient_bn_relu = nn.Sequential(
            nn.BatchNorm2d(out_channel),
            nn.ReLU(inplace=True)
        )

        #"""The form of the shortcut connection can be an identity function with
        #the appropriate padding or a projection as in He et al. (2015b)."""
        self.short_cut = nn.Sequential()
        if in_channel != out_channel or stride != 1:
            self.short_cut = nn.Sequential(
                nn.Conv2d(in_channel, out_channel, kernel_size=1, stride=stride)
            )


    def forward(self, x):
        x_residual, x_transient = x
        residual_r_r = self.residual_stream_conv(x_residual)
        residual_r_t = self.residual_stream_conv_across(x_residual)
        residual_shortcut = self.short_cut(x_residual)

        transient_t_t = self.transient_stream_conv(x_transient)
        transient_t_r = self.transient_stream_conv_across(x_transient)

        #transient_t_t = self.transient_stream_conv(x_residual)
        #transient_t_r = self.transient_stream_conv_across(x_residual)
        #"""Same-stream and cross-stream activations are summed (along with the
        #shortcut connection for the residual stream) before applying batch
        #normalization and ReLU nonlinearities (together σ) to get the output
        #states of the block (Equation 1) (Ioffe & Szegedy, 2015)."""
        x_residual = self.residual_bn_relu(residual_r_r + transient_t_r + residual_shortcut)
        x_transient = self.transient_bn_relu(residual_r_t + transient_t_t)

        return x_residual, x_transient



class RiRBlock(nn.Module):
    def __init__(self, in_channel, out_channel, layer_num, stride, layer=ResnetInit):
        super().__init__()
        self.resnetinit = self._make_layers(in_channel, out_channel, layer_num, stride)

        #self.short_cut = nn.Sequential()
        #if stride != 1 or in_channel != out_channel:
        #    self.short_cut = nn.Conv2d(in_channel, out_channel, kernel_size=1, stride=stride)

    def forward(self, x):
        x_residual, x_transient = self.resnetinit(x)
        #x_residual = x_residual + self.short_cut(x[0])
        #x_transient = x_transient + self.short_cut(x[1])

        return (x_residual, x_transient)

    #"""Replacing each of the convolutional layers within a residual
    #block from the original ResNet (Figure 1a) with a generalized residual block
    #(Figure 1b) leads us to a new architecture we call ResNet in ResNet (RiR)
    #(Figure 1d)."""
    def _make_layers(self, in_channel, out_channel, layer_num, stride, layer=ResnetInit):
        strides = [stride] + [1] * (layer_num - 1)
        layers = nn.Sequential()
        for index, s in enumerate(strides):
            layers.add_module("generalized layers{}".format(index), layer(in_channel, out_channel, s))
            in_channel = out_channel

        return layers

class ResnetInResneet(nn.Module):
    def __init__(self, num_classes=100):
        super().__init__()
        base = int(96 / 2)
        self.residual_pre_conv = nn.Sequential(
            nn.Conv2d(3, base, 3, padding=1),
            nn.BatchNorm2d(base),
            nn.ReLU(inplace=True)
        )
        self.transient_pre_conv = nn.Sequential(
            nn.Conv2d(3, base, 3, padding=1),
            nn.BatchNorm2d(base),
            nn.ReLU(inplace=True)
        )

        self.rir1 = RiRBlock(base, base, 2, 1)
        self.rir2 = RiRBlock(base, base, 2, 1)
        self.rir3 = RiRBlock(base, base * 2, 2, 2)
        self.rir4 = RiRBlock(base * 2, base * 2, 2, 1)
        self.rir5 = RiRBlock(base * 2, base * 2, 2, 1)
        self.rir6 = RiRBlock(base * 2, base * 4, 2, 2)
        self.rir7 = RiRBlock(base * 4, base * 4, 2, 1)
        self.rir8 = RiRBlock(base * 4, base * 4, 2, 1)

        self.conv1 = nn.Sequential(
            nn.Conv2d(384, num_classes, kernel_size=3, stride=2), #without this convolution, loss will soon be nan
            nn.BatchNorm2d(num_classes),
            nn.ReLU(inplace=True),
        )

        self.classifier = nn.Sequential(
            nn.Linear(900, 450),
            nn.ReLU(),
            nn.Dropout(),
            nn.Linear(450, 100),
        )

        self._weight_init()

    def forward(self, x):
        x_residual = self.residual_pre_conv(x)
        x_transient = self.transient_pre_conv(x)

        x_residual, x_transient = self.rir1((x_residual, x_transient))
        x_residual, x_transient = self.rir2((x_residual, x_transient))
        x_residual, x_transient = self.rir3((x_residual, x_transient))
        x_residual, x_transient = self.rir4((x_residual, x_transient))
        x_residual, x_transient = self.rir5((x_residual, x_transient))
        x_residual, x_transient = self.rir6((x_residual, x_transient))
        x_residual, x_transient = self.rir7((x_residual, x_transient))
        x_residual, x_transient = self.rir8((x_residual, x_transient))
        h = torch.cat([x_residual, x_transient], 1)
        h = self.conv1(h)
        h = h.view(h.size()[0], -1)
        h = self.classifier(h)

        return h

    def _weight_init(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                torch.nn.init.kaiming_normal(m.weight)
                m.bias.data.fill_(0.01)


def resnet_in_resnet():
    return ResnetInResneet()



"""mobilenet in pytorch

[1] Andrew G. Howard, Menglong Zhu, Bo Chen, Dmitry Kalenichenko, Weijun Wang, Tobias Weyand, Marco Andreetto, Hartwig Adam

    MobileNets: Efficient Convolutional Neural Networks for Mobile Vision Applications
    https://arxiv.org/abs/1704.04861
"""

import torch
import torch.nn as nn


class DepthSeperabelConv1d(nn.Module):

    def __init__(self, input_channels, output_channels, kernel_size, **kwargs):
        super().__init__()
        self.depthwise = nn.Sequential(
            nn.Conv1d(
                input_channels,
                input_channels,
                kernel_size,
                groups=input_channels,
                **kwargs),
            nn.BatchNorm1d(input_channels),
            nn.ReLU(inplace=True)
        )

        self.pointwise = nn.Sequential(
            nn.Conv1d(input_channels, output_channels, 1),
            nn.BatchNorm1d(output_channels),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        x = self.depthwise(x)
        x = self.pointwise(x)

        return x


class BasicConv1d(nn.Module):

    def __init__(self, input_channels, output_channels, kernel_size, **kwargs):
        super().__init__()
        self.conv = nn.Conv1d(
            input_channels, output_channels, kernel_size, **kwargs)
        self.bn = nn.BatchNorm1d(output_channels)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)

        return x


class MobileNet(nn.Module):
    """
    Args:
        width multipler: The role of the width multiplier α is to thin
                         a network uniformly at each layer. For a given
                         layer and width multiplier α, the number of
                         input channels M becomes αM and the number of
                         output channels N becomes αN.
    """

    def __init__(self, width_multiplier=1, class_num=1):
        super().__init__()

        alpha = width_multiplier
        self.stem = nn.Sequential(
            BasicConv1d(1, int(32 * alpha), 3, padding=1, bias=False),
            DepthSeperabelConv1d(
                int(32 * alpha),
                int(64 * alpha),
                3,
                padding=1,
                bias=False
            )
        )

        #downsample
        self.conv1 = nn.Sequential(
            DepthSeperabelConv1d(
                int(64 * alpha),
                int(128 * alpha),
                3,
                stride=2,
                padding=1,
                bias=False
            ),
            DepthSeperabelConv1d(
                int(128 * alpha),
                int(128 * alpha),
                3,
                padding=1,
                bias=False
            )
        )

        #downsample
        self.conv2 = nn.Sequential(
            DepthSeperabelConv1d(
                int(128 * alpha),
                int(256 * alpha),
                3,
                stride=2,
                padding=1,
                bias=False
            ),
            DepthSeperabelConv1d(
                int(256 * alpha),
                int(256 * alpha),
                3,
                padding=1,
                bias=False
            )
        )

        #downsample
        self.conv3 = nn.Sequential(
            DepthSeperabelConv1d(
                int(256 * alpha),
                int(512 * alpha),
                3,
                stride=2,
                padding=1,
                bias=False
            ),

            DepthSeperabelConv1d(
                int(512 * alpha),
                int(512 * alpha),
                3,
                padding=1,
                bias=False
            ),
            DepthSeperabelConv1d(
                int(512 * alpha),
                int(512 * alpha),
                3,
                padding=1,
                bias=False
            ),
            DepthSeperabelConv1d(
                int(512 * alpha),
                int(512 * alpha),
                3,
                padding=1,
                bias=False
            ),
            DepthSeperabelConv1d(
                int(512 * alpha),
                int(512 * alpha),
                3,
                padding=1,
                bias=False
            ),
            DepthSeperabelConv1d(
                int(512 * alpha),
                int(512 * alpha),
                3,
                padding=1,
                bias=False
            )
        )

        #downsample
        self.conv4 = nn.Sequential(
            DepthSeperabelConv1d(
                int(512 * alpha),
                int(1024 * alpha),
                3,
                stride=2,
                padding=1,
                bias=False
            ),
            DepthSeperabelConv1d(
                int(1024 * alpha),
                int(1024 * alpha),
                3,
                padding=1,
                bias=False
            )
        )

        self.fc = nn.Linear(int(1024 * alpha), class_num)
        self.avg = nn.AdaptiveAvgPool1d(1)

    def forward(self, x):
        x = self.stem(x)

        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)

        x = self.avg(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        return x


def mobilenet(alpha=1, class_num=2):
    return MobileNet(alpha, class_num)



""" inceptionv3 in pytorch


[1] Christian Szegedy, Vincent Vanhoucke, Sergey Ioffe, Jonathon Shlens, Zbigniew Wojna

    Rethinking the Inception Architecture for Computer Vision
    https://arxiv.org/abs/1512.00567v3
"""

import torch
import torch.nn as nn


class BasicConv2d(nn.Module):

    def __init__(self, input_channels, output_channels, **kwargs):
        super().__init__()
        self.conv = nn.Conv2d(input_channels, output_channels, bias=False, **kwargs)
        self.bn = nn.BatchNorm2d(output_channels)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)

        return x

#same naive inception module
class InceptionA(nn.Module):

    def __init__(self, input_channels, pool_features):
        super().__init__()
        self.branch1x1 = BasicConv2d(input_channels, 64, kernel_size=1)

        self.branch5x5 = nn.Sequential(
            BasicConv2d(input_channels, 48, kernel_size=1),
            BasicConv2d(48, 64, kernel_size=5, padding=2)
        )

        self.branch3x3 = nn.Sequential(
            BasicConv2d(input_channels, 64, kernel_size=1),
            BasicConv2d(64, 96, kernel_size=3, padding=1),
            BasicConv2d(96, 96, kernel_size=3, padding=1)
        )

        self.branchpool = nn.Sequential(
            nn.AvgPool2d(kernel_size=3, stride=1, padding=1),
            BasicConv2d(input_channels, pool_features, kernel_size=3, padding=1)
        )

    def forward(self, x):

        #x -> 1x1(same)
        branch1x1 = self.branch1x1(x)

        #x -> 1x1 -> 5x5(same)
        branch5x5 = self.branch5x5(x)
        #branch5x5 = self.branch5x5_2(branch5x5)

        #x -> 1x1 -> 3x3 -> 3x3(same)
        branch3x3 = self.branch3x3(x)

        #x -> pool -> 1x1(same)
        branchpool = self.branchpool(x)

        outputs = [branch1x1, branch5x5, branch3x3, branchpool]

        return torch.cat(outputs, 1)

#downsample
#Factorization into smaller convolutions
class InceptionB(nn.Module):

    def __init__(self, input_channels):
        super().__init__()

        self.branch3x3 = BasicConv2d(input_channels, 384, kernel_size=3, stride=2)

        self.branch3x3stack = nn.Sequential(
            BasicConv2d(input_channels, 64, kernel_size=1),
            BasicConv2d(64, 96, kernel_size=3, padding=1),
            BasicConv2d(96, 96, kernel_size=3, stride=2)
        )

        self.branchpool = nn.MaxPool2d(kernel_size=3, stride=2)

    def forward(self, x):

        #x - > 3x3(downsample)
        branch3x3 = self.branch3x3(x)

        #x -> 3x3 -> 3x3(downsample)
        branch3x3stack = self.branch3x3stack(x)

        #x -> avgpool(downsample)
        branchpool = self.branchpool(x)

        #"""We can use two parallel stride 2 blocks: P and C. P is a pooling
        #layer (either average or maximum pooling) the activation, both of
        #them are stride 2 the filter banks of which are concatenated as in
        #figure 10."""
        outputs = [branch3x3, branch3x3stack, branchpool]

        return torch.cat(outputs, 1)

#Factorizing Convolutions with Large Filter Size
class InceptionC(nn.Module):
    def __init__(self, input_channels, channels_7x7):
        super().__init__()
        self.branch1x1 = BasicConv2d(input_channels, 192, kernel_size=1)

        c7 = channels_7x7

        #In theory, we could go even further and argue that one can replace any n × n
        #convolution by a 1 × n convolution followed by a n × 1 convolution and the
        #computational cost saving increases dramatically as n grows (see figure 6).
        self.branch7x7 = nn.Sequential(
            BasicConv2d(input_channels, c7, kernel_size=1),
            BasicConv2d(c7, c7, kernel_size=(7, 1), padding=(3, 0)),
            BasicConv2d(c7, 192, kernel_size=(1, 7), padding=(0, 3))
        )

        self.branch7x7stack = nn.Sequential(
            BasicConv2d(input_channels, c7, kernel_size=1),
            BasicConv2d(c7, c7, kernel_size=(7, 1), padding=(3, 0)),
            BasicConv2d(c7, c7, kernel_size=(1, 7), padding=(0, 3)),
            BasicConv2d(c7, c7, kernel_size=(7, 1), padding=(3, 0)),
            BasicConv2d(c7, 192, kernel_size=(1, 7), padding=(0, 3))
        )

        self.branch_pool = nn.Sequential(
            nn.AvgPool2d(kernel_size=3, stride=1, padding=1),
            BasicConv2d(input_channels, 192, kernel_size=1),
        )

    def forward(self, x):

        #x -> 1x1(same)
        branch1x1 = self.branch1x1(x)

        #x -> 1layer 1*7 and 7*1 (same)
        branch7x7 = self.branch7x7(x)

        #x-> 2layer 1*7 and 7*1(same)
        branch7x7stack = self.branch7x7stack(x)

        #x-> avgpool (same)
        branchpool = self.branch_pool(x)

        outputs = [branch1x1, branch7x7, branch7x7stack, branchpool]

        return torch.cat(outputs, 1)

class InceptionD(nn.Module):

    def __init__(self, input_channels):
        super().__init__()

        self.branch3x3 = nn.Sequential(
            BasicConv2d(input_channels, 192, kernel_size=1),
            BasicConv2d(192, 320, kernel_size=3, stride=2)
        )

        self.branch7x7 = nn.Sequential(
            BasicConv2d(input_channels, 192, kernel_size=1),
            BasicConv2d(192, 192, kernel_size=(1, 7), padding=(0, 3)),
            BasicConv2d(192, 192, kernel_size=(7, 1), padding=(3, 0)),
            BasicConv2d(192, 192, kernel_size=3, stride=2)
        )

        self.branchpool = nn.AvgPool2d(kernel_size=3, stride=2)

    def forward(self, x):

        #x -> 1x1 -> 3x3(downsample)
        branch3x3 = self.branch3x3(x)

        #x -> 1x1 -> 1x7 -> 7x1 -> 3x3 (downsample)
        branch7x7 = self.branch7x7(x)

        #x -> avgpool (downsample)
        branchpool = self.branchpool(x)

        outputs = [branch3x3, branch7x7, branchpool]

        return torch.cat(outputs, 1)


#same
class InceptionE(nn.Module):
    def __init__(self, input_channels):
        super().__init__()
        self.branch1x1 = BasicConv2d(input_channels, 320, kernel_size=1)

        self.branch3x3_1 = BasicConv2d(input_channels, 384, kernel_size=1)
        self.branch3x3_2a = BasicConv2d(384, 384, kernel_size=(1, 3), padding=(0, 1))
        self.branch3x3_2b = BasicConv2d(384, 384, kernel_size=(3, 1), padding=(1, 0))

        self.branch3x3stack_1 = BasicConv2d(input_channels, 448, kernel_size=1)
        self.branch3x3stack_2 = BasicConv2d(448, 384, kernel_size=3, padding=1)
        self.branch3x3stack_3a = BasicConv2d(384, 384, kernel_size=(1, 3), padding=(0, 1))
        self.branch3x3stack_3b = BasicConv2d(384, 384, kernel_size=(3, 1), padding=(1, 0))

        self.branch_pool = nn.Sequential(
            nn.AvgPool2d(kernel_size=3, stride=1, padding=1),
            BasicConv2d(input_channels, 192, kernel_size=1)
        )

    def forward(self, x):

        #x -> 1x1 (same)
        branch1x1 = self.branch1x1(x)

        # x -> 1x1 -> 3x1
        # x -> 1x1 -> 1x3
        # concatenate(3x1, 1x3)
        #"""7. Inception modules with expanded the filter bank outputs.
        #This architecture is used on the coarsest (8 × 8) grids to promote
        #high dimensional representations, as suggested by principle
        #2 of Section 2."""
        branch3x3 = self.branch3x3_1(x)
        branch3x3 = [
            self.branch3x3_2a(branch3x3),
            self.branch3x3_2b(branch3x3)
        ]
        branch3x3 = torch.cat(branch3x3, 1)

        # x -> 1x1 -> 3x3 -> 1x3
        # x -> 1x1 -> 3x3 -> 3x1
        #concatenate(1x3, 3x1)
        branch3x3stack = self.branch3x3stack_1(x)
        branch3x3stack = self.branch3x3stack_2(branch3x3stack)
        branch3x3stack = [
            self.branch3x3stack_3a(branch3x3stack),
            self.branch3x3stack_3b(branch3x3stack)
        ]
        branch3x3stack = torch.cat(branch3x3stack, 1)

        branchpool = self.branch_pool(x)

        outputs = [branch1x1, branch3x3, branch3x3stack, branchpool]

        return torch.cat(outputs, 1)

class InceptionV3(nn.Module):

    def __init__(self, num_classes=100):
        super().__init__()
        self.Conv2d_1a_3x3 = BasicConv2d(3, 32, kernel_size=3, padding=1)
        self.Conv2d_2a_3x3 = BasicConv2d(32, 32, kernel_size=3, padding=1)
        self.Conv2d_2b_3x3 = BasicConv2d(32, 64, kernel_size=3, padding=1)
        self.Conv2d_3b_1x1 = BasicConv2d(64, 80, kernel_size=1)
        self.Conv2d_4a_3x3 = BasicConv2d(80, 192, kernel_size=3)

        #naive inception module
        self.Mixed_5b = InceptionA(192, pool_features=32)
        self.Mixed_5c = InceptionA(256, pool_features=64)
        self.Mixed_5d = InceptionA(288, pool_features=64)

        #downsample
        self.Mixed_6a = InceptionB(288)

        self.Mixed_6b = InceptionC(768, channels_7x7=128)
        self.Mixed_6c = InceptionC(768, channels_7x7=160)
        self.Mixed_6d = InceptionC(768, channels_7x7=160)
        self.Mixed_6e = InceptionC(768, channels_7x7=192)

        #downsample
        self.Mixed_7a = InceptionD(768)

        self.Mixed_7b = InceptionE(1280)
        self.Mixed_7c = InceptionE(2048)

        #6*6 feature size
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.dropout = nn.Dropout2d()
        self.linear = nn.Linear(2048, num_classes)

    def forward(self, x):

        #32 -> 30
        x = self.Conv2d_1a_3x3(x)
        x = self.Conv2d_2a_3x3(x)
        x = self.Conv2d_2b_3x3(x)
        x = self.Conv2d_3b_1x1(x)
        x = self.Conv2d_4a_3x3(x)

        #30 -> 30
        x = self.Mixed_5b(x)
        x = self.Mixed_5c(x)
        x = self.Mixed_5d(x)

        #30 -> 14
        #Efficient Grid Size Reduction to avoid representation
        #bottleneck
        x = self.Mixed_6a(x)

        #14 -> 14
        #"""In practice, we have found that employing this factorization does not
        #work well on early layers, but it gives very good results on medium
        #grid-sizes (On m × m feature maps, where m ranges between 12 and 20).
        #On that level, very good results can be achieved by using 1 × 7 convolutions
        #followed by 7 × 1 convolutions."""
        x = self.Mixed_6b(x)
        x = self.Mixed_6c(x)
        x = self.Mixed_6d(x)
        x = self.Mixed_6e(x)

        #14 -> 6
        #Efficient Grid Size Reduction
        x = self.Mixed_7a(x)

        #6 -> 6
        #We are using this solution only on the coarsest grid,
        #since that is the place where producing high dimensional
        #sparse representation is the most critical as the ratio of
        #local processing (by 1 × 1 convolutions) is increased compared
        #to the spatial aggregation."""
        x = self.Mixed_7b(x)
        x = self.Mixed_7c(x)

        #6 -> 1
        x = self.avgpool(x)
        x = self.dropout(x)
        x = x.view(x.size(0), -1)
        x = self.linear(x)
        return x


def inceptionv3():
    return InceptionV3()






import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Optional


class UNet(nn.Module):
    def __init__(self, n_channels: int, n_classes: int, bilinear: bool = False) -> None:
        super(UNet, self).__init__()
        self.n_channels = n_channels
        self.n_classes = n_classes
        self.bilinear = bilinear

        self.inc = DoubleConv(n_channels, 64)
        self.down1 = Down(64, 128)
        self.down2 = Down(128, 256)
        self.down3 = Down(256, 512)
        factor = 2 if bilinear else 1
        self.down4 = Down(512, 1024 // factor)
        self.up1 = Up(1024, 512 // factor, bilinear)
        self.up2 = Up(512, 256 // factor, bilinear)
        self.up3 = Up(256, 128 // factor, bilinear)
        self.up4 = Up(128, 64, bilinear)
        self.outc = OutConv(64, n_classes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        logits = self.outc(x)
        return logits

    def use_checkpointing(self) -> None:
        self.inc = torch.utils.checkpoint(self.inc)
        self.down1 = torch.utils.checkpoint(self.down1)
        self.down2 = torch.utils.checkpoint(self.down2)
        self.down3 = torch.utils.checkpoint(self.down3)
        self.down4 = torch.utils.checkpoint(self.down4)
        self.up1 = torch.utils.checkpoint(self.up1)
        self.up2 = torch.utils.checkpoint(self.up2)
        self.up3 = torch.utils.checkpoint(self.up3)
        self.up4 = torch.utils.checkpoint(self.up4)
        self.outc = torch.utils.checkpoint(self.outc)


class DoubleConv(nn.Module):
    """(convolution => [BN] => ReLU) * 2"""

    def __init__(self, in_channels: int, out_channels: int, mid_channels: Optional[int] = None) -> None:
        super(DoubleConv, self).__init__()
        if not mid_channels:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            nn.Conv1d(in_channels, mid_channels, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm1d(mid_channels),
            nn.ReLU(inplace=True),
            nn.Conv1d(mid_channels, out_channels, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm1d(out_channels),
            nn.ReLU(inplace=True)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.double_conv(x)


class Down(nn.Module):
    """Downscaling with maxpool then double conv"""

    def __init__(self, in_channels: int, out_channels: int) -> None:
        super(Down, self).__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool1d(2),
            DoubleConv(in_channels, out_channels)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.maxpool_conv(x)


class Up(nn.Module):
    """Upscaling then double conv"""

    def __init__(self, in_channels: int, out_channels: int, bilinear: bool = True) -> None:
        super(Up, self).__init__()

        if bilinear:
            self.up = nn.Upsample(scale_factor=2, mode='linear', align_corners=True)
            self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)
        else:
            self.up = nn.ConvTranspose1d(in_channels, in_channels // 2, kernel_size=2, stride=2)
            self.conv = DoubleConv(in_channels, out_channels)

    def forward(self, x1: torch.Tensor, x2: torch.Tensor) -> torch.Tensor:
        x1 = self.up(x1)
        diffX = x2.size()[2] - x1.size()[2]
        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2])
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)


class OutConv(nn.Module):
    def __init__(self, in_channels: int, out_channels: int) -> None:
        super(OutConv, self).__init__()
        self.conv = nn.Conv1d(in_channels, out_channels, kernel_size=1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.conv(x)



"""senet in pytorch



[1] Jie Hu, Li Shen, Samuel Albanie, Gang Sun, Enhua Wu

    Squeeze-and-Excitation Networks
    https://arxiv.org/abs/1709.01507
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

class BasicResidualSEBlock(nn.Module):

    expansion = 1

    def __init__(self, in_channels, out_channels, stride, r=16):
        super().__init__()

        self.residual = nn.Sequential(
            nn.Conv1d(in_channels, out_channels, 3, stride=stride, padding=1),
            nn.BatchNorm1d(out_channels),
            nn.ReLU(inplace=True),

            nn.Conv1d(out_channels, out_channels * self.expansion, 3, padding=1),
            nn.BatchNorm1d(out_channels * self.expansion),
            nn.ReLU(inplace=True)
        )

        self.shortcut = nn.Sequential()
        if stride != 1 or in_channels != out_channels * self.expansion:
            self.shortcut = nn.Sequential(
                nn.Conv1d(in_channels, out_channels * self.expansion, 1, stride=stride),
                nn.BatchNorm1d(out_channels * self.expansion)
            )

        self.squeeze = nn.AdaptiveAvgPool1d(1)
        self.excitation = nn.Sequential(
            nn.Linear(out_channels * self.expansion, out_channels * self.expansion // r),
            nn.ReLU(inplace=True),
            nn.Linear(out_channels * self.expansion // r, out_channels * self.expansion),
            nn.Sigmoid()
        )

    def forward(self, x):
        shortcut = self.shortcut(x)
        residual = self.residual(x)

        squeeze = self.squeeze(residual)
        squeeze = squeeze.view(squeeze.size(0), -1)
        excitation = self.excitation(squeeze)
        excitation = excitation.view(residual.size(0), residual.size(1), 1)

        x = residual * excitation.expand_as(residual) + shortcut

        return F.relu(x)

class BottleneckResidualSEBlock(nn.Module):

    expansion = 4

    def __init__(self, in_channels, out_channels, stride, r=16):
        super().__init__()

        self.residual = nn.Sequential(
            nn.Conv1d(in_channels, out_channels, 1),
            nn.BatchNorm1d(out_channels),
            nn.ReLU(inplace=True),

            nn.Conv1d(out_channels, out_channels, 3, stride=stride, padding=1),
            nn.BatchNorm1d(out_channels),
            nn.ReLU(inplace=True),

            nn.Conv1d(out_channels, out_channels * self.expansion, 1),
            nn.BatchNorm1d(out_channels * self.expansion),
            nn.ReLU(inplace=True)
        )

        self.squeeze = nn.AdaptiveAvgPool1d(1)
        self.excitation = nn.Sequential(
            nn.Linear(out_channels * self.expansion, out_channels * self.expansion // r),
            nn.ReLU(inplace=True),
            nn.Linear(out_channels * self.expansion // r, out_channels * self.expansion),
            nn.Sigmoid()
        )

        self.shortcut = nn.Sequential()
        if stride != 1 or in_channels != out_channels * self.expansion:
            self.shortcut = nn.Sequential(
                nn.Conv1d(in_channels, out_channels * self.expansion, 1, stride=stride),
                nn.BatchNorm1d(out_channels * self.expansion)
            )

    def forward(self, x):

        shortcut = self.shortcut(x)

        residual = self.residual(x)
        squeeze = self.squeeze(residual)
        squeeze = squeeze.view(squeeze.size(0), -1)
        excitation = self.excitation(squeeze)
        excitation = excitation.view(residual.size(0), residual.size(1), 1)

        x = residual * excitation.expand_as(residual) + shortcut

        return F.relu(x)

class SEResNet(nn.Module):

    def __init__(self, block, block_num, class_num=2):
        super().__init__()

        self.in_channels = 64

        self.pre = nn.Sequential(
            nn.Conv1d(1, 64, 3, padding=1),
            nn.BatchNorm1d(64),
            nn.ReLU(inplace=True)
        )

        self.stage1 = self._make_stage(block, block_num[0], 64, 1)
        self.stage2 = self._make_stage(block, block_num[1], 128, 2)
        self.stage3 = self._make_stage(block, block_num[2], 256, 2)
        self.stage4 = self._make_stage(block, block_num[3], 512, 2)

        self.linear = nn.Linear(self.in_channels, class_num)

    def forward(self, x):
        b_size = x.shape[0]
        x = self.pre(x)

        x = self.stage1(x)
        x = self.stage2(x)
        x = self.stage3(x)
        x = self.stage4(x)

        x = F.adaptive_avg_pool1d(x, 1)
        x = x.view(x.size(0), -1)

        x = self.linear(x)
        x =x.reshape(b_size, 1, -1)
        return x


    def _make_stage(self, block, num, out_channels, stride):

        layers = []
        layers.append(block(self.in_channels, out_channels, stride))
        self.in_channels = out_channels * block.expansion

        while num - 1:
            layers.append(block(self.in_channels, out_channels, 1))
            num -= 1

        return nn.Sequential(*layers)

def seresnet18():
    return SEResNet(BasicResidualSEBlock, [2, 2, 2, 2])

def seresnet34():
    return SEResNet(BasicResidualSEBlock, [3, 4, 6, 3])

def seresnet50():
    return SEResNet(BottleneckResidualSEBlock, [3, 4, 6, 3])

def seresnet101():
    return SEResNet(BottleneckResidualSEBlock, [3, 4, 23, 3])

def seresnet152():
    return SEResNet(BottleneckResidualSEBlock, [3, 8, 36, 3])



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Here are 4 easy steps to use Fabric in your PyTorch code.

1. Create the Lightning Fabric object at the beginning of your script.

2. Remove all ``.to`` and ``.cuda`` calls since Fabric will take care of it.

3. Apply ``setup`` over each model and optimizers pair, ``setup_dataloaders`` on all your dataloaders,
and replace ``loss.backward()`` with ``self.backward(loss)``.

4. Run the script from the terminal using ``fabric run path/to/train.py``

Accelerate your training loop by setting the ``--accelerator``, ``--strategy``, ``--devices`` options directly from
the command line. See ``fabric run --help`` or learn more from the documentation:
https://lightning.ai/docs/fabric.

"""

import argparse
from os import path

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torchvision.transforms as T
from lightning.fabric import Fabric, seed_everything
from torch.optim.lr_scheduler import StepLR
from torchmetrics.classification import Accuracy
from torchvision.datasets import MNIST

DATASETS_PATH = "./tmp/data1/MNIST"


class Net(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.conv1 = nn.Conv2d(1, 32, 3, 1)
        self.conv2 = nn.Conv2d(32, 64, 3, 1)
        self.dropout1 = nn.Dropout(0.25)
        self.dropout2 = nn.Dropout(0.5)
        self.fc1 = nn.Linear(9216, 128)
        self.fc2 = nn.Linear(128, 10)

    def forward(self, x):
        x = self.conv1(x)
        x = F.relu(x)
        x = self.conv2(x)
        x = F.relu(x)
        x = F.max_pool2d(x, 2)
        x = self.dropout1(x)
        x = torch.flatten(x, 1)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.dropout2(x)
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)


def run(hparams):
    # Create the Lightning Fabric object. The parameters like accelerator, strategy, devices etc. will be proided
    # by the command line. See all options: `fabric run --help`
    fabric = Fabric()

    seed_everything(hparams.seed)  # instead of torch.manual_seed(...)

    transform = T.Compose([T.ToTensor(), T.Normalize((0.1307,), (0.3081,))])

    # Let rank 0 download the data first, then everyone will load MNIST
    with fabric.rank_zero_first(local=False):  # set `local=True` if your filesystem is not shared between machines
        train_dataset = MNIST(DATASETS_PATH, download=fabric.is_global_zero, train=True, transform=transform)
        test_dataset = MNIST(DATASETS_PATH, download=fabric.is_global_zero, train=False, transform=transform)

    train_loader = torch.utils.data.DataLoader(
        train_dataset,
        batch_size=hparams.batch_size,
    )
    test_loader = torch.utils.data.DataLoader(test_dataset, batch_size=hparams.batch_size)

    # don't forget to call `setup_dataloaders` to prepare for dataloaders for distributed training.
    train_loader, test_loader = fabric.setup_dataloaders(train_loader, test_loader)

    model = Net()  # remove call to .to(device)
    optimizer = optim.Adadelta(model.parameters(), lr=hparams.lr)

    # don't forget to call `setup` to prepare for model / optimizer for distributed training.
    # the model is moved automatically to the right device.
    model, optimizer = fabric.setup(model, optimizer)

    scheduler = StepLR(optimizer, step_size=1, gamma=hparams.gamma)

    # use torchmetrics instead of manually computing the accuracy
    test_acc = Accuracy(task="multiclass", num_classes=10).to(fabric.device)

    # EPOCH LOOP
    for epoch in range(1, hparams.epochs + 1):
        # TRAINING LOOP
        model.train()
        for batch_idx, (data, target) in enumerate(train_loader):
            # NOTE: no need to call `.to(device)` on the data, target
            optimizer.zero_grad()
            output = model(data)
            loss = F.nll_loss(output, target)
            fabric.backward(loss)  # instead of loss.backward()

            optimizer.step()
            if (batch_idx == 0) or ((batch_idx + 1) % hparams.log_interval == 0):
                print(
                    f"Train Epoch: {epoch} [{batch_idx * len(data)}/{len(train_loader.dataset)}"
                    f" ({100.0 * batch_idx / len(train_loader):.0f}%)]\tLoss: {loss.item():.6f}"
                )
                if hparams.dry_run:
                    break

        scheduler.step()

        # TESTING LOOP
        model.eval()
        test_loss = 0
        with torch.no_grad():
            for data, target in test_loader:
                # NOTE: no need to call `.to(device)` on the data, target
                output = model(data)
                test_loss += F.nll_loss(output, target, reduction="sum").item()

                # WITHOUT TorchMetrics
                # pred = output.argmax(dim=1, keepdim=True)  # get the index of the max log-probability
                # correct += pred.eq(target.view_as(pred)).sum().item()

                # WITH TorchMetrics
                test_acc(output, target)

                if hparams.dry_run:
                    break

        # all_gather is used to aggregated the value across processes
        test_loss = fabric.all_gather(test_loss).sum() / len(test_loader.dataset)

        print(f"\nTest set: Average loss: {test_loss:.4f}, Accuracy: ({100 * test_acc.compute():.0f}%)\n")
        test_acc.reset()

        if hparams.dry_run:
            break

    # When using distributed training, use `fabric.save`
    # to ensure the current process is allowed to save a checkpoint
    if hparams.save_model:
        fabric.save(model.state_dict(), "mnist_cnn.pt")


if __name__ == "__main__":
    # Arguments can be passed in through the CLI as normal and will be parsed here
    # Example:
    # fabric run image_classifier.py accelerator=cuda --epochs=3
    parser = argparse.ArgumentParser(description="Fabric MNIST Example")
    parser.add_argument(
        "--batch-size", type=int, default=64, metavar="N", help="input batch size for training (default: 64)"
    )
    parser.add_argument("--epochs", type=int, default=14, metavar="N", help="number of epochs to train (default: 14)")
    parser.add_argument("--lr", type=float, default=1.0, metavar="LR", help="learning rate (default: 1.0)")
    parser.add_argument("--gamma", type=float, default=0.7, metavar="M", help="Learning rate step gamma (default: 0.7)")
    parser.add_argument("--dry-run", action="store_true", default=False, help="quickly check a single pass")
    parser.add_argument("--seed", type=int, default=1, metavar="S", help="random seed (default: 1)")
    parser.add_argument(
        "--log-interval",
        type=int,
        default=10,
        metavar="N",
        help="how many batches to wait before logging training status",
    )
    parser.add_argument("--save-model", action="store_true", default=False, help="For Saving the current Model")
    hparams = parser.parse_args()

    run(hparams)


from typing import Any, Callable, Dict, List, Optional, Sequence, Union

import torch
from torch import nn


class Compose:
    """Composes several transforms together.

    This transform does not support torchscript.
    Please, see the note below.

    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.

    Example:
        >>> transforms.Compose([
        >>>     transforms.CenterCrop(10),
        >>>     transforms.PILToTensor(),
        >>>     transforms.ConvertImageDtype(torch.float),
        >>> ])

    .. note::
        In order to script the transformations, please use ``torch.nn.Sequential`` as below.

        >>> transforms = torch.nn.Sequential(
        >>>     transforms.CenterCrop(10),
        >>>     transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
        >>> )
        >>> scripted_transforms = torch.jit.script(transforms)

        Make sure to use only scriptable transformations, i.e. that work with ``torch.Tensor``, does not require
        `lambda` functions or ``PIL.Image``.

    """

    def __init__(self, transforms: Sequence[Callable]) -> None:
        super().__init__()
        if not isinstance(transforms, Sequence):
            raise TypeError("Argument transforms should be a sequence of callables")
        elif not transforms:
            raise ValueError("Pass at least one transform")
        self.transforms = transforms

    def __call__(self, *inputs: Any) -> Any:
        needs_unpacking = len(inputs) > 1
        for transform in self.transforms:
            outputs = transform(*inputs)
            inputs = outputs if needs_unpacking else (outputs,)
        return outputs

    def __len__(self):
        return len(self.transforms)

    def extra_repr(self) -> str:
        format_string = []
        for t in self.transforms:
            format_string.append(f"    {t}")
        return "\n".join(format_string)


class RandomApply():
    """Apply randomly a list of transformations with a given probability.

    .. note::
        In order to script the transformation, please use ``torch.nn.ModuleList`` as input instead of list/tuple of
        transforms as shown below:

        >>> transforms = transforms.RandomApply(torch.nn.ModuleList([
        >>>     transforms.ColorJitter(),
        >>> ]), p=0.3)
        >>> scripted_transforms = torch.jit.script(transforms)

        Make sure to use only scriptable transformations, i.e. that work with ``torch.Tensor``, does not require
        `lambda` functions or ``PIL.Image``.

    Args:
        transforms (sequence or torch.nn.Module): list of transformations
        p (float): probability of applying the list of transforms
    """

    def __init__(self, transforms: Union[Sequence[Callable], nn.ModuleList], p: float = 0.5) -> None:
        super().__init__()

        if not isinstance(transforms, (Sequence, nn.ModuleList)):
            raise TypeError("Argument transforms should be a sequence of callables or a `nn.ModuleList`")
        self.transforms = transforms

        if not (0.0 <= p <= 1.0):
            raise ValueError("`p` should be a floating point value in the interval [0.0, 1.0].")
        self.p = p

    def _extract_params_for_v1_transform(self) -> Dict[str, Any]:
        return {"transforms": self.transforms, "p": self.p}

    def forward(self, *inputs: Any) -> Any:
        needs_unpacking = len(inputs) > 1

        if torch.rand(1) >= self.p:
            return inputs if needs_unpacking else inputs[0]

        for transform in self.transforms:
            outputs = transform(*inputs)
            inputs = outputs if needs_unpacking else (outputs,)
        return outputs

    def extra_repr(self) -> str:
        format_string = []
        for t in self.transforms:
            format_string.append(f"    {t}")
        return "\n".join(format_string)


class RandomChoice:
    """Apply single transformation randomly picked from a list.

    This transform does not support torchscript.

    Args:
        transforms (sequence or torch.nn.Module): list of transformations
        p (list of floats or None, optional): probability of each transform being picked.
            If ``p`` doesn't sum to 1, it is automatically normalized. If ``None``
            (default), all transforms have the same probability.
    """

    def __init__(
            self,
            transforms: Sequence[Callable],
            p: Optional[List[float]] = None,
    ) -> None:
        if not isinstance(transforms, Sequence):
            raise TypeError("Argument transforms should be a sequence of callables")

        if p is None:
            p = [1] * len(transforms)
        elif len(p) != len(transforms):
            raise ValueError(f"Length of p doesn't match the number of transforms: {len(p)} != {len(transforms)}")

        super().__init__()

        self.transforms = transforms
        total = sum(p)
        self.p = [prob / total for prob in p]

    def forward(self, *inputs: Any) -> Any:
        idx = int(torch.multinomial(torch.tensor(self.p), 1))
        transform = self.transforms[idx]
        return transform(*inputs)


class RandomOrder:
    """Apply a list of transformations in a random order.

    This transform does not support torchscript.

    Args:
        transforms (sequence or torch.nn.Module): list of transformations
    """

    def __init__(self, transforms: Sequence[Callable]) -> None:
        if not isinstance(transforms, Sequence):
            raise TypeError("Argument transforms should be a sequence of callables")
        super().__init__()
        self.transforms = transforms

    def forward(self, *inputs: Any) -> Any:
        needs_unpacking = len(inputs) > 1
        for idx in torch.randperm(len(self.transforms)):
            transform = self.transforms[idx]
            outputs = transform(*inputs)
            inputs = outputs if needs_unpacking else (outputs,)
        return outputs



from typing import List

import torch
import torch.nn.functional as F
# from models import BaseVAE
from torch import nn, Tensor
# from torch import tensor as Tensor
from torch.nn import SmoothL1Loss

map_size = 10  # length after final conventional  20*32


class DepthSeparableConv1d(nn.Module):

    def __init__(self, input_channels, out_channels, kernel_size, **kwargs):
        super().__init__()
        self.Depthwise = nn.Sequential(
            nn.Conv1d(
                input_channels, input_channels,
                kernel_size, groups=input_channels, **kwargs),
            nn.BatchNorm1d(input_channels),
            nn.LeakyReLU(inplace=True)
        )

        self.Pointwise = nn.Sequential(
            nn.Conv1d(input_channels, out_channels, 1),
            nn.BatchNorm1d(out_channels),
            nn.LeakyReLU(inplace=True)
        )

    def forward(self, x):
        x = self.Depthwise(x)
        x = self.Pointwise(x)
        return x


class BasicConv1d(nn.Module):

    def __init__(self, input_channels, out_channels, kernel_size, **kwargs):
        super().__init__()
        self.conv = nn.Conv1d(
            input_channels, out_channels, kernel_size, **kwargs)
        self.bn = nn.BatchNorm1d(out_channels)
        self.relu = nn.LeakyReLU()

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)

        return x


class DepthSeparableConvTranspose1d(nn.Module):

    def __init__(self, input_channels, out_channels, kernel_size, **kwargs):
        super().__init__()
        self.Depthwise = nn.Sequential(
            nn.ConvTranspose1d(input_channels, input_channels,
                               kernel_size, groups=input_channels, **kwargs),
            nn.BatchNorm1d(input_channels),
            nn.LeakyReLU(inplace=True)
        )

        self.Pointwise = nn.Sequential(
            nn.Conv1d(input_channels, out_channels, 1),
            nn.BatchNorm1d(out_channels),
            nn.LeakyReLU(inplace=True)
        )

    def forward(self, x):
        x = self.Depthwise(x)
        # print("dw:", x.size())
        x = self.Pointwise(x)
        # print("pw:", x.size())
        # print("--------------")
        return x


class Hswish(nn.Module):
    def __init__(self, inplace=True):
        super(Hswish, self).__init__()
        self.inplace = inplace

    def forward(self, x):
        return x * F.relu6(x + 3., inplace=self.inplace) / 6.


class Hsigmoid(nn.Module):
    def __init__(self, inplace=True):
        super(Hsigmoid, self).__init__()
        self.inplace = inplace

    def forward(self, x):
        return F.relu6(x + 3., inplace=self.inplace) / 6.


class SEModule(nn.Module):
    def __init__(self, channel, reduction=4):
        super(SEModule, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool1d(1)
        self.fc = nn.Sequential(
            nn.Linear(channel, channel // reduction, bias=False),
            nn.ReLU(inplace=True),
            nn.Linear(channel // reduction, channel, bias=False),
            Hsigmoid()
            # nn.Sigmoid()
        )

    def forward(self, x):
        b, c, _ = x.size()
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1)
        return x * y.expand_as(x)


class Identity(nn.Module):
    def __init__(self, channel):
        super(Identity, self).__init__()

    def forward(self, x):
        return x


def make_divisible(x, divisible_by=8):
    import numpy as np
    return int(np.ceil(x * 1. / divisible_by) * divisible_by)


class MobileBottleneck(nn.Module):
    def __init__(self, in_channels, oup, kernel, stride, exp, se=False, nl='RE'):
        super(MobileBottleneck, self).__init__()
        assert stride in [1, 2]
        assert kernel in [3, 5]
        padding = (kernel - 1) // 2
        self.use_res_connect = stride == 1 and in_channels == oup

        conv_layer = nn.Conv1d
        norm_layer = nn.BatchNorm1d
        if nl == 'RE':
            nlin_layer = nn.ReLU  # or ReLU6
        elif nl == 'HS':
            nlin_layer = Hswish
        else:
            raise NotImplementedError
        if se:
            SELayer = SEModule
        else:
            SELayer = Identity

        self.conv = nn.Sequential(
            # pw
            conv_layer(in_channels, exp, 1, 1, 0, bias=False),
            norm_layer(exp),
            nlin_layer(inplace=True),
            # dw
            conv_layer(exp, exp, kernel, stride, padding, groups=exp, bias=False),
            norm_layer(exp),
            SELayer(exp),
            nlin_layer(inplace=True),
            # pw-linear
            conv_layer(exp, oup, 1, 1, 0, bias=False),
            norm_layer(oup),
        )

    def forward(self, x):
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)


class MobileNetV3(nn.Module):

    def __init__(self,
                 in_channels: int,
                 latent_dim: int,
                 hidden_dims: List = None,
                 **kwargs) -> None:
        super().__init__()

        self.latent_dim = latent_dim

        modules_i = []
        if hidden_dims is None:
            hidden_dims = [32, 64, 128, 256, 512, 1024]

        # Build Encoder 通过Sequential将网络层和激活函数结合起来，输出激活后的五层网络节点
        modules_i.append(
            nn.Sequential(
                BasicConv1d(in_channels, out_channels=32,
                            kernel_size=3, stride=2, padding=1),
                MobileBottleneck(32, 64, 3, 2, exp=88, se=False, nl='RE'),
                MobileBottleneck(64, 128, 3, 2, exp=96, se=True, nl='RE'),
                MobileBottleneck(128, 256, 3, 2, exp=288, se=True, nl='HS'),
                MobileBottleneck(256, 512, 3, 2, exp=512, se=True, nl='HS'),
                MobileBottleneck(512, 1024, 3, 2, exp=1024, se=True, nl='HS')

            ))
        self.encoder_i = nn.Sequential(*modules_i)  # 将输入i加载到网络上

        self.fc_mu_i = nn.Sequential(
            nn.Linear(hidden_dims[-1] * map_size, latent_dim),  # 将1024*20转化为64维,
            nn.BatchNorm1d(latent_dim),
            nn.LeakyReLU())

        # Build Decoder
        modules = []
        self.decoder_input = nn.Sequential(
            nn.Linear(latent_dim, hidden_dims[-1] * map_size),  # 将64维,
            nn.BatchNorm1d(hidden_dims[-1] * map_size),
            nn.LeakyReLU())

        hidden_dims.reverse()

        for i in range(len(hidden_dims) - 1):
            modules.append(
                nn.Sequential(
                    DepthSeparableConvTranspose1d(hidden_dims[i], hidden_dims[i + 1],
                                                  kernel_size=3, stride=2, padding=1, output_padding=1)))
        self.decoder = nn.Sequential(*modules)

        self.final_layer = nn.Sequential(
            DepthSeparableConvTranspose1d(hidden_dims[-1], hidden_dims[-1],
                                          kernel_size=3, stride=2, padding=1, output_padding=1),
            nn.Conv1d(hidden_dims[-1], out_channels=1,
                      kernel_size=3, padding=1),
            nn.Sigmoid()
        )

    def encode_i(self, input: Tensor) -> List[Tensor]:
        """
        Encodes the input by passing through the encoder network
        and returns the latent codes.
        :param input: (Tensor) Input tensor to encoder [N x C x H x W]
        :return: (Tensor) List of latent codes
        """
        result = self.encoder_i(input)
        # print(result.shape)
        # flatten it to 1 dim to better performance
        result = torch.flatten(result, start_dim=1)
        # print(result.shape)
        # with the latent Gaussian distribution
        mu_i = self.fc_mu_i(result)
        # print(mu_i.shape)
        return mu_i

    def decode(self, z: Tensor) -> Tensor:
        """
        Maps the given latent codes
        onto the image space.
        :param z: (Tensor) [B x D]
        :return: (Tensor) [B x C x H x W]
        """
        result = self.decoder_input(z)
        # print(result.shape)
        result = result.view(-1, 1024, map_size)
        # print(result.shape)
        result = self.decoder(result)
        result = self.final_layer(result)
        return result

    def loss_function(self, *args, gt) -> dict:
        recons = args[0]
        recons_loss_ = SmoothL1Loss()
        recons_loss = recons_loss_(recons.flatten(), gt.flatten())
        loss = recons_loss
        return loss

    def forward(self, input_i: Tensor, **kwargs) -> List[Tensor]:

        mu_i = self.encode_i(input_i)

        return [self.decode(mu_i), input_i, mu_i]

    def sample(self,
               num_samples: int,
               current_device: int, **kwargs) -> Tensor:
        """
        Samples from the latent space and return the corresponding
        image space map.
        :param num_samples: (Int) Number of samples
        :param current_device: (Int) Device to run the models
        :return: (Tensor)
        """
        z = torch.randn(num_samples,
                        self.latent_dim)
        z = z.to(current_device)

        samples = self.decode(z)
        return samples

    def generate(self, x: Tensor, **kwargs) -> Tensor:
        """
        Given an input image x, returns the reconstructed image
        :param x: (Tensor) [B x C x H x W]
        :return: (Tensor) [B x C x H x W]
        """

        return self.forward(x)[0]



import torch.nn as nn
from utils import show_parameter


class Encoder(nn.Module):
    def __init__(self,
                 in_channels: int,
                 hidden_dims: list,
                 latent_dim: int,
                 map_size: int,
                 length: int):
        super(Encoder, self).__init__()
        modules = []
        for h_dim in hidden_dims:
            length = length / 2
            modules.append(nn.Sequential(nn.Conv1d(in_channels, out_channels=h_dim,
                                                   kernel_size=3, stride=2, padding=1),
                                         nn.LayerNorm(int(length)),
                                         nn.LeakyReLU()))
            in_channels = h_dim
        self.net = nn.Sequential(*modules)
        self.fc = nn.Sequential(nn.Flatten(),
                                nn.Linear(hidden_dims[-1]
                                          * map_size, latent_dim),
                                nn.LeakyReLU())

    def forward(self, x):
        x = self.net(x)
        x = self.fc(x)
        return x


class Decoder(nn.Module):
    def __init__(self,
                 in_channels: int,
                 hidden_dims: list,
                 latent_dim: int,
                 map_size: int,
                 length: int):
        super(Decoder, self).__init__()
        modules = []
        self.map_size = map_size
        hidden_dims.reverse()
        self.hidden_dims = hidden_dims
        length = map_size
        for i in range(len(hidden_dims) - 1):
            length = 2 * length
            modules.append(
                nn.Sequential(nn.ConvTranspose1d(hidden_dims[i], hidden_dims[i + 1],
                                                 kernel_size=3, stride=2, padding=1,
                                                 output_padding=1),
                              nn.LayerNorm(int(length)),
                              nn.LeakyReLU())
            )
        self.net = nn.Sequential(*modules)
        self.fc = nn.Sequential(nn.Linear(latent_dim, hidden_dims[0] * map_size),
                                nn.LeakyReLU())

    def forward(self, x):
        x = self.fc(x)
        # print(x)
        x = x.reshape(-1, self.hidden_dims[0], self.map_size)
        x = self.net(x)
        return x


class Autoencoder(nn.Module):
    def __init__(
            self,
            num_input_channels: int,
            latent_dim: int,
            hidden_dims: list = None,
            length: int = 3520,
    ):
        super().__init__()

        if hidden_dims is None:
            hidden_dims = [32, 64, 128, 256]
        feature_map_size = int(length / pow(2, len(hidden_dims)))

        # Creating encoder and decoder
        self.encoder = Encoder(
            num_input_channels, hidden_dims, latent_dim, feature_map_size, length)
        self.decoder = Decoder(
            num_input_channels, hidden_dims, latent_dim, feature_map_size, length)
        self.final_layer = nn.Sequential(
            nn.Conv1d(hidden_dims[-1],
                               hidden_dims[-1],
                               kernel_size=3,
                               stride=2,
                               padding=1,),
            nn.BatchNorm1d(hidden_dims[-1]),
            nn.LeakyReLU(),
            nn.Conv1d(hidden_dims[-1], out_channels=1,
                      kernel_size=3, padding=1),
            nn.Linear(880,440),
            nn.Dropout(0.5),
            nn.Linear(440,2)
        )

    def forward(self, x):
        """The forward function takes in an image and returns the reconstructed image."""
        z = self.encoder(x)
        x_hat = self.decoder(z)
        x_hat = self.final_layer(x_hat)
        return x_hat

    def get_reconstruction_loss(self, *output, labels):
        """Given a batch of images, this function returns the reconstruction loss (MSE in our case)"""
        x_hat = output[0]
        _loss = nn.SmoothL1Loss()
        loss = _loss(x_hat, labels)
        return loss


if __name__ == "__main__":
    model = Autoencoder(1, 128).cuda()
    show_parameter(model)



import torch
import torch.nn as nn
# from models.MobileNet.MobileNetV1 import DepthSeparableConv1d, DepthSeparableConvTranspose1d
import torch.nn.functional as F
import os
# classes
import yaml
from torch.autograd import Variable
from thop import profile
from thop import clever_format


class Encoder(nn.Module):
    def __init__(self,
                 in_channels: int,
                 hidden_dims: list,
                 latent_dim: int,
                 map_size: int,
                 length: int):
        super(Encoder, self).__init__()
        # modules = [nn.Dropout(0.2)]
        modules = []
        for h_dim in hidden_dims:
            length = length / 2
            modules.append(nn.Sequential(nn.Conv1d(in_channels, out_channels=h_dim,
                                                   kernel_size=3, stride=2, padding=1),
                                         nn.LayerNorm(int(length)),
                                         nn.LeakyReLU()))
            in_channels = h_dim
        self.net = nn.Sequential(*modules)
        self.fc = nn.Sequential(nn.Flatten(),
                                nn.Linear(hidden_dims[-1] * map_size, latent_dim),
                                nn.LeakyReLU())

    def forward(self, x):
        x = self.net(x)
        # x = self.fc(x)
        return x


class Autoencoder(nn.Module):
    def __init__(
            self,
            num_input_channels: int,
            latent_dim: int,
            hidden_dims: list = None,
            length: int = 3520,
    ):
        super().__init__()
        if hidden_dims is None:
            hidden_dims = [32, 64,128,256]
        feature_map_size = int(length / pow(2, len(hidden_dims)))

        # Creating encoder and decoder
        self.encoder = Encoder(num_input_channels, hidden_dims, latent_dim, feature_map_size, length)
        self.final_layer2 = nn.Sequential(
            nn.Linear(latent_dim, 256),
            nn.Sigmoid(),
            nn.Linear(256, 2)
        )
        self.final_layer = nn.Sequential(nn.Conv1d(256,128,
                                               kernel_size=3,
                                               stride=2,
                                               padding=1),
                            nn.BatchNorm1d(128),
                            nn.LeakyReLU(),
                            nn.Conv1d(128, out_channels= 1,
                                      kernel_size= 3, padding= 1),
                            nn.Dropout(0.2),
                            nn.Linear(110,1),
                           )
    def forward(self, x):
        """The forward function takes in an image and returns the reconstructed image."""

        b_size = x.shape[0]
        z = self.encoder(x)
        x_hat = self.final_layer(z)
        x_hat = x_hat.reshape(b_size, 1, -1)
        return x_hat

if __name__ == '__main__':
    # gradient check
    batch_size = 1
    model = Autoencoder(1,64).cuda()
    loss_fn = torch.nn.MSELoss()
    input = Variable(torch.randn(batch_size, 1, 3520)).cuda()
    target = Variable(torch.randn(batch_size, 1, 2)).double().cuda()
    flops, params = profile(model, inputs=(input, ))
    flops, params = clever_format([flops, params], "%.3f")
    output = model(input)
    output = output[0].double()
    res = torch.autograd.gradcheck(loss_fn, (output.flatten(), target.flatten()), eps=1e-6, raise_exception=True)
    print(flops, params)
    print(output.size())


import torch
import numpy.random as random
from .utils import *


class Interpolate(torch.nn.Module):
    def __init__(self, size=None, scale_factor=None, mode='linear'):
        """
        :param min_snr: Minimum signal-to-noise ratio
        :param max_snr: Maximum signal-to-noise ratio
        """
        super().__init__()
        self.size = size
        self.scale_factor = scale_factor
        self.mode = mode

    def forward(self, x):
        if not has_batch_dimension(x):
            x = add_batch_dimension(x)
        x = torch.nn.functional.interpolate(x, size=self.size, mode=self.mode)
        x = remove_batch_dimension(x)
        return x



from typing import Any, Callable, cast, Dict, List, Optional, Tuple, Union
import numpy as np
import torch
from h5py import File


def load_file(file_path, keys):
    file_extension = file_path.split('.')[-1].lower()

    if file_extension == '.txt':
        return load_text_file(file_path)
    elif file_extension == 'csv':
        return load_csv_file(file_path)
    elif file_extension == 'json':
        return load_json_file(file_path)
    elif file_extension == 'mat':
        return load_mat_file(file_path, keys)
    elif file_extension == 'npz':
        return load_npz_file(file_path, keys)
    else:
        raise ValueError("Unsupported file format")


def load_text_file(file_path):
    with open(file_path, 'r') as file:
        return file.read()


def load_csv_file(file_path):
    import csv
    data = []
    with open(file_path, 'r') as file:
        csv_reader = csv.reader(file)
        for row in csv_reader:
            data.append(row)
    return data


def load_json_file(file_path):
    import json
    with open(file_path, 'r') as file:
        return json.load(file)


def load_mat_file(path, list_j):
    from scipy import io
    try:
        all_mat_data = io.loadmat(path)
        return get_meta_data(all_mat_data, list_j)

    except NotImplementedError:
        all_mat_data = File(path)
        return get_meta_data(all_mat_data, list_j)


def get_meta_data(
        raw_data: Union[Dict[str, Any], File],
        keys: Tuple[List[str], Optional[List[str]]]
) -> Tuple:
    def parse_key(key: str) -> Tuple[str, Optional[int]]:
        if '[' in key and ']' in key:
            key_name = key[:key.index('[')]
            row_str = key[key.index('[') + 1:key.index(']')]
            if row_str.isdigit():
                return key_name, int(row_str)
        return key, None

    if keys is None:
        # If keys is None, return all values in raw_data as a tuple
        # Remove meta info
        if '__header__' in raw_data:
            del raw_data['__header__']
        if '__version__' in raw_data:
            del raw_data['__version__']
        if '__globals__' in raw_data:
            del raw_data['__globals__']
        return tuple(raw_data.values())
    else:
        inputs, targets = keys
        input_values = []
        target_values = []

        # 提取inputs并处理特定行
        for key in inputs:
            key_name, row = parse_key(key)
            if key_name in raw_data:
                value = raw_data[key_name]
                if row is not None and len(value) > row:
                    input_values.append(value[row])
                else:
                    input_values.append(value)

        # 提取targets
        for key in targets:
            key_name, row = parse_key(key)
            if key_name in raw_data:
                value = raw_data[key_name]
                if row is not None and len(value) > row:
                    target_values.append(value[row])
                else:
                    target_values.append(value)

        # 将列表转换为元组
        return input_values, target_values


def load_npz_file(path, keys):
    npz_data = np.load(path)
    return tuple(npz_data[key] for key in keys if key in npz_data)



# 使用示例
# file_path = 'example.txt'
# data1 = load_file(file_path)
# print(data1)



import os
import os.path
from pathlib import Path
from typing import Any, Callable, cast, Dict, List, Optional, Tuple, Union

from torch.utils.data import random_split
from .vision import VisionDataset


def has_file_allowed_extension(filename: str, extensions: Union[str, Tuple[str, ...]]) -> bool:
    """Checks if a file is an allowed extension and does not contain 'move' in the filename.

    Args:
        filename (string): path to a file
        extensions (tuple of strings): extensions to consider (lowercase)

    Returns:
        bool: True if the filename ends with one of given extensions and does not contain 'move'
    """
    return 'move' not in filename.lower() and filename.lower().endswith(
        extensions if isinstance(extensions, str) else tuple(extensions))


def is_image_file(filename: str) -> bool:
    """Checks if a file is an allowed image extension.

    Args:
        filename (string): path to a file

    Returns:
        bool: True if the filename ends with a known image extension
    """
    return has_file_allowed_extension(filename, IMG_EXTENSIONS)


def find_classes(directory: Union[str, Path]) -> Tuple[List[str], Dict[str, int]]:
    """Finds the class folders in a dataset.

    See :class:`DatasetFolder` for details.
    """
    classes = sorted(entry.name for entry in os.scandir(directory) if entry.is_dir())
    if not classes:
        raise FileNotFoundError(f"Couldn't find any class folder in {directory}.")

    class_to_idx = {cls_name: i for i, cls_name in enumerate(classes)}
    return classes, class_to_idx


def make_dataset(
        directory: Union[str, Path],
        class_to_idx: Optional[Dict[str, int]] = None,
        extensions: Optional[Union[str, Tuple[str, ...]]] = None,
        is_valid_file: Optional[Callable[[str], bool]] = None,
        allow_empty: bool = False,
) -> List[Tuple[str, int]]:
    """Generates a list of samples of a form (path_to_sample, class).

    See :class:`DatasetFolder` for details.

    Note: The class_to_idx parameter is here optional and will use the logic of the ``find_classes`` function
    by default.
    """
    directory = os.path.expanduser(directory)

    if class_to_idx is None:
        _, class_to_idx = find_classes(directory)
    elif not class_to_idx:
        raise ValueError("'class_to_index' must have at least one entry to collect any samples.")

    both_none = extensions is None and is_valid_file is None
    both_something = extensions is not None and is_valid_file is not None
    if both_none or both_something:
        raise ValueError("Both extensions and is_valid_file cannot be None or not None at the same time")

    if extensions is not None:
        def is_valid_file(x: str) -> bool:
            return has_file_allowed_extension(x, extensions)  # type: ignore[arg-type]

    is_valid_file = cast(Callable[[str], bool], is_valid_file)

    instances = []
    available_classes = set()
    for target_class in sorted(class_to_idx.keys()):
        class_index = class_to_idx[target_class]
        target_dir = os.path.join(directory, target_class)
        if not os.path.isdir(target_dir):
            continue
        for root, _, fnames in sorted(os.walk(target_dir, followlinks=True)):
            for fname in sorted(fnames):
                path = os.path.join(root, fname)
                # print(path)
                if is_valid_file(path):
                    item = path, class_index
                    instances.append(item)

                    if target_class not in available_classes:
                        available_classes.add(target_class)

    empty_classes = set(class_to_idx.keys()) - available_classes
    if empty_classes and not allow_empty:
        msg = f"Found no valid file for the classes {', '.join(sorted(empty_classes))}. "
        if extensions is not None:
            msg += f"Supported extensions are: {extensions if isinstance(extensions, str) else ', '.join(extensions)}"
        raise FileNotFoundError(msg)

    return instances


class DatasetFolder(VisionDataset):
    """A generic data1 loader.

    This default directory structure can be customized by overriding the
    :meth:`find_classes` method.

    Args:
        root (str or ``pathlib.Path``): Root directory path.
        loader (callable): A function to load a sample given its path.
        extensions (tuple[string]): A list of allowed extensions.
            both extensions and is_valid_file should not be passed.
        transform (callable, optional): A function/transform that takes in
            a sample and returns a transformed version.
            E.g, ``transforms.RandomCrop`` for images.
        target_transform (callable, optional): A function/transform that takes
            in the target and transforms it.
        is_valid_file (callable, optional): A function that takes path of a file
            and check if the file is a valid file (used to check of corrupt files)
            both extensions and is_valid_file should not be passed.
        allow_empty(bool, optional): If True, empty folders are considered to be valid classes.
            An error is raised on empty folders if False (default).

     Attributes:
        classes (list): List of the class names sorted alphabetically.
        class_to_idx (dict): Dict with items (class_name, class_index).
        samples (list): List of (sample path, class_index) tuples
        targets (list): The class_index value for each image in the dataset
    """

    def __init__(
            self,
            root: Union[str, Path],
            params: Tuple[List[str], Optional[List[str]]],
            loader: Callable[[str, Any], Any],
            extensions: Optional[Tuple[str, ...]] = None,
            transform: Union[Optional[Callable], List[Callable]] = None,
            target_transform: Optional[Callable] = None,
            is_valid_file: Optional[Callable[[str], bool]] = None,
            allow_empty: bool = False,
    ) -> None:
        super().__init__(root, transform=transform, target_transform=target_transform)
        classes, class_to_idx = self.find_classes(self.root)
        samples = self.make_dataset(
            self.root,
            class_to_idx=class_to_idx,
            extensions=extensions,
            is_valid_file=is_valid_file,
            allow_empty=allow_empty,
        )
        self.params = params
        self.loader = loader
        self.extensions = extensions

        self.classes = classes
        self.class_to_idx = class_to_idx
        self.samples = samples
        self.targets = [s[1] for s in samples]

    @staticmethod
    def make_dataset(
            directory: Union[str, Path],
            class_to_idx: Dict[str, int],
            extensions: Optional[Tuple[str, ...]] = None,
            is_valid_file: Optional[Callable[[str], bool]] = None,
            allow_empty: bool = False,
    ) -> List[Tuple[str, int]]:
        """Generates a list of samples of a form (path_to_sample, class).

        This can be overridden to e.g. read files from a compressed zip file instead of from the disk.

        Args:
            directory (str): root dataset directory, corresponding to ``self.root``.
            class_to_idx (Dict[str, int]): Dictionary mapping class name to class index.
            extensions (optional): A list of allowed extensions.
                Either extensions or is_valid_file should be passed. Defaults to None.
            is_valid_file (optional): A function that takes path of a file
                and checks if the file is a valid file
                (used to check of corrupt files) both extensions and
                is_valid_file should not be passed. Defaults to None.
            allow_empty(bool, optional): If True, empty folders are considered to be valid classes.
                An error is raised on empty folders if False (default).

        Raises:
            ValueError: In case ``class_to_idx`` is empty.
            ValueError: In case ``extensions`` and ``is_valid_file`` are None or both are not None.
            FileNotFoundError: In case no valid file was found for any class.

        Returns:
            List[Tuple[str, int]]: samples of a form (path_to_sample, class)
        """
        if class_to_idx is None:
            # prevent potential bug since make_dataset() would use the class_to_idx logic of the
            # find_classes() function, instead of using that of the find_classes() method, which
            # is potentially overridden and thus could have a different logic.
            raise ValueError("The class_to_idx parameter cannot be None.")
        return make_dataset(
            directory, class_to_idx, extensions=extensions, is_valid_file=is_valid_file, allow_empty=allow_empty
        )

    def find_classes(self, directory: Union[str, Path]) -> Tuple[List[str], Dict[str, int]]:
        """Find the class folders in a dataset structured as follows::

            directory/
            ├── class_x
            │   ├── xxx.ext
            │   ├── xxy.ext
            │   └── ...
            │       └── xxz.ext
            └── class_y
                ├── 123.ext
                ├── nsdf3.ext
                └── ...
                └── asd932_.ext

        This method can be overridden to only consider
        a subset of classes, or to adapt to a different dataset directory structure.

        Args:
            directory(str): Root directory path, corresponding to ``self.root``

        Raises:
            FileNotFoundError: If ``dir`` has no class folders.

        Returns:
            (Tuple[List[str], Dict[str, int]]): List of all classes and dictionary mapping each class to an index.
        """
        return find_classes(directory)

    def __getitem__(self, index: int) -> Tuple[Any, Any, Any]:
        """
        Args:
            index (int): Index

        Returns:
            tuple: (sample, target) where target is class_index of the target class.
        """
        path, labels = self.samples[index]
        sample, target = self.loader(path, self.params)
        if len(sample) == 1:
            sample = sample[0]
        if len(target) == 1:
            target = target[0]
        if self.transform is not None:
            for i in range(len(self.transform)):
                sample[i] = self.transform[i](sample[i])
        if self.target_transform is not None:
            target = self.target_transform(target)

        return sample, target, labels

    def __len__(self) -> int:
        return len(self.samples)

    def train_test_split(self, val_ratio):
        """
        划分训练集和验证集的函数，接收一个参数：`val_ratio`（验证集的比例）。
        这个函数会返回划分好的训练集和验证集。
        """
        val_size = int(val_ratio * len(self))
        train_size = len(self) - val_size
        train_set, val_set = random_split(self, [train_size, val_size])

        return train_set, val_set


IMG_EXTENSIONS = (".mat", ".jpeg", ".npz")



import torch
from torch import nn, optim
import numpy as np
from einops import rearrange, repeat, pack, unpack
from einops.layers.torch import Rearrange
import lightning.pytorch as pl


# classes

class PreNorm(nn.Module):
    def __init__(self, dim, fn):
        super().__init__()
        self.norm = nn.LayerNorm(dim)
        self.fn = fn

    def forward(self, x, **kwargs):
        return self.fn(self.norm(x), **kwargs)


class FeedForward(nn.Module):
    def __init__(self, dim, hidden_dim, dropout=0.):
        super().__init__()
        self.net = nn.Sequential(
            nn.Conv1d(dim, hidden_dim, kernel_size=3, padding=1),
            nn.LeakyReLU(),
            nn.Dropout(dropout),
            nn.Conv1d(hidden_dim, dim, kernel_size=3, padding=1),
            nn.Dropout(dropout)
        )

    def forward(self, x):
        # Reshape input tensor to fit the expected shape of Conv1d
        x = x.transpose(1, 2)  # Assuming x has shape (batch_size, dim, sequence_length)

        x = self.net(x)
        x = x.transpose(1, 2)  # Restore the original shape
        # print(x.shape)
        return x


class Attention(nn.Module):
    def __init__(self, dim, heads=8, dim_head=64, dropout=0.):
        super().__init__()
        inner_dim = dim_head * heads
        project_out = not (heads == 1 and dim_head == dim)

        self.heads = heads
        self.scale = dim_head ** -0.5

        self.attend = nn.Softmax(dim=-1)
        self.dropout = nn.Dropout(dropout)

        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias=False)

        self.to_out = nn.Sequential(
            nn.Linear(inner_dim, dim),
            nn.Dropout(dropout)
        ) if project_out else nn.Identity()

    def forward(self, x):
        qkv = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h=self.heads), qkv)

        dots = torch.matmul(q, k.transpose(-1, -2)) * self.scale

        attn = self.attend(dots)
        attn = self.dropout(attn)

        out = torch.matmul(attn, v)
        out = rearrange(out, 'b h n d -> b n (h d)')
        return self.to_out(out)


class Transformer(nn.Module):
    def __init__(self, dim, depth, heads, dim_head, mlp_dim, dropout=0.):
        super().__init__()
        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers.append(nn.ModuleList([
                PreNorm(dim, Attention(dim, heads=heads, dim_head=dim_head, dropout=dropout)),
                PreNorm(dim, FeedForward(dim, mlp_dim, dropout=dropout))
            ]))

    def forward(self, x):
        for attn, ff in self.layers:
            x = attn(x) + x
            # print(x.shape)
            x = ff(x) + x
            # print(x.shape)
        return x


class CosineWarmupScheduler(optim.lr_scheduler._LRScheduler):
    def __init__(self, optimizer, warmup, max_iters):
        self.warmup = warmup
        self.max_num_iters = max_iters
        super().__init__(optimizer)

    def get_lr(self):
        lr_factor = self.get_lr_factor(epoch=self.last_epoch)
        return [base_lr * lr_factor for base_lr in self.base_lrs]

    def get_lr_factor(self, epoch):
        lr_factor = 0.5 * (1 + np.cos(np.pi * epoch / self.max_num_iters))
        if epoch <= self.warmup:
            lr_factor *= epoch * 1.0 / self.warmup
        return lr_factor


class ViT(pl.LightningModule):
    def __init__(self, configs):
        super().__init__()
        seq_len = configs["model_params"]["seq_len"]
        patch_size = configs["model_params"]["patch_size"]
        num_classes = configs["model_params"]["num_classes"]
        dim = configs["model_params"]["dim"]
        depth = configs["model_params"]["depth"]
        heads = configs["model_params"]["heads"]
        mlp_dim = configs["model_params"]["mlp_dim"]
        dropout = configs["model_params"]["dropout"]
        emb_dropout = configs["model_params"]["emb_dropout"]
        channels = configs["model_params"]["in_channels"]
        dim_head = configs["model_params"]["dim_head"]
        assert (seq_len % patch_size) == 0
        self.save_hyperparameters()
        self.training_step_outputs = []
        self.val_step_outputs = []
        num_patches = seq_len // patch_size
        patch_dim = channels * patch_size

        self.to_patch_embedding = nn.Sequential(
            Rearrange('b c (n p) -> b n (p c)', p=patch_size),
            nn.LayerNorm(patch_dim),
            nn.Linear(patch_dim, dim),
            nn.LayerNorm(dim),
        )

        self.pos_embedding = nn.Parameter(torch.randn(1, num_patches + 1, dim))
        self.cls_token = nn.Parameter(torch.randn(dim))
        self.dropout = nn.Dropout(emb_dropout)

        self.transformer = Transformer(dim, depth, heads, dim_head, mlp_dim, dropout)

        self.mlp_head = nn.Sequential(
            nn.LayerNorm(dim),
            nn.Linear(dim, num_classes)
        )

    def forward(self, series):
        x = self.to_patch_embedding(series)
        b, n, _ = x.shape

        cls_tokens = repeat(self.cls_token, 'd -> b d', b=b)

        x, ps = pack([cls_tokens, x], 'b * d')

        x += self.pos_embedding[:, :(n + 1)]
        x = self.dropout(x)

        x = self.transformer(x)

        cls_tokens, _ = unpack(x, ps, 'b * d')

        return self.mlp_head(cls_tokens)

    def _get_reconstruction_loss(self, batch):
        x, labels, _ = batch
        x_hat = self.forward(x)
        _loss = nn.SmoothL1Loss()
        loss = _loss(x_hat.flatten(), labels.flatten())
        return loss

    def configure_optimizers(self):
        optimizer = optim.AdamW(self.parameters(), lr=1e-4)
        self.lr_scheduler = CosineWarmupScheduler(
            optimizer, warmup=5, max_iters=200
        )
        return optimizer

    def optimizer_step(self, *args, **kwargs):
        super().optimizer_step(*args, **kwargs)
        self.lr_scheduler.step()  # Step per iteration

    def training_step(self, train_batch, batch_idx):
        loss = self._get_reconstruction_loss(train_batch)
        self.training_step_outputs.append(loss)
        self.log("train_loss", loss)
        # print(loss)
        return loss

    def validation_step(self, val_batch, batch_idx):
        loss = self._get_reconstruction_loss(val_batch)
        self.val_step_outputs.append(loss)
        self.log("val_loss", loss, "batch_size", batch_size=50)
        return loss


if __name__ == '__main__':
    v = ViT(
        seq_len=3520,
        patch_size=64,
        num_classes=1,
        dim=512,
        depth=6,
        heads=8,
        mlp_dim=2048,
        dropout=0.2,
        emb_dropout=0.1
    )

    time_series = torch.randn(32, 1, 3520)
    logits = v(time_series)  # (4, 1000)
    print(logits.shape)



import torch
from torch import Tensor
from typing import Callable, Optional, Sequence, Tuple, Union
from torchaudio import functional as F


class GriffinLim(torch.nn.Module):
    r"""Compute waveform from a linear scale magnitude spectrogram using the Griffin-Lim transformation.

    .. devices:: CPU CUDA

    .. properties:: Autograd TorchScript

    Implementation ported from
    *librosa* :cite:`brian_mcfee-proc-scipy-2015`, *A fast Griffin-Lim algorithm* :cite:`6701851`
    and *Signal estimation from modified short-time Fourier transform* :cite:`1172092`.

    Args:
        n_fft (int, optional): Size of FFT, creates ``n_fft // 2 + 1`` bins. (Default: ``400``)
        n_iter (int, optional): Number of iteration for phase recovery process. (Default: ``32``)
        win_length (int or None, optional): Window size. (Default: ``n_fft``)
        hop_length (int or None, optional): Length of hop between STFT windows. (Default: ``win_length // 2``)
        window_fn (Callable[..., Tensor], optional): A function to create a window tensor
            that is applied/multiplied to each frame/window. (Default: ``torch.hann_window``)
        power (float, optional): Exponent for the magnitude spectrogram,
            (must be > 0) e.g., 1 for magnitude, 2 for power, etc. (Default: ``2``)
        wkwargs (dict or None, optional): Arguments for window function. (Default: ``None``)
        momentum (float, optional): The momentum parameter for fast Griffin-Lim.
            Setting this to 0 recovers the original Griffin-Lim method.
            Values near 1 can lead to faster convergence, but above 1 may not converge. (Default: ``0.99``)
        length (int, optional): Array length of the expected output. (Default: ``None``)
        rand_init (bool, optional): Initializes phase randomly if True and to zero otherwise. (Default: ``True``)

    Example
        >>> batch, freq, time = 2, 257, 100
        >>> spectrogram = torch.randn(batch, freq, time)
        >>> transform = transforms.GriffinLim(n_fft=512)
        >>> waveform = transform(spectrogram)
    """
    __constants__ = ["n_fft", "n_iter", "win_length", "hop_length", "power", "length", "momentum", "rand_init"]

    def __init__(
            self,
            n_fft: int = 400,
            n_iter: int = 32,
            win_length: Optional[int] = None,
            hop_length: Optional[int] = None,
            window_fn: Callable[..., Tensor] = torch.hann_window,
            power: float = 2.0,
            wkwargs: Optional[dict] = None,
            momentum: float = 0.99,
            length: Optional[int] = None,
            rand_init: bool = True,
    ) -> None:
        super(GriffinLim, self).__init__()

        if not (0 <= momentum < 1):
            raise ValueError("momentum must be in the range [0, 1). Found: {}".format(momentum))

        self.n_fft = n_fft
        self.n_iter = n_iter
        self.win_length = win_length if win_length is not None else n_fft
        self.hop_length = hop_length if hop_length is not None else self.win_length // 2
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        window = window_fn(self.win_length) if wkwargs is None else window_fn(self.win_length, **wkwargs)
        self.register_buffer("window", window.to(device))
        self.length = length
        self.power = power
        self.momentum = momentum
        self.rand_init = rand_init

    def forward(self, specgram: Tensor) -> Tensor:
        r"""
        Args:
            specgram (Tensor):
                A magnitude-only STFT spectrogram of dimension (..., freq, frames)
                where freq is ``n_fft // 2 + 1``.

        Returns:
            Tensor: waveform of (..., time), where time equals the ``length`` parameter if given.
        """
        return F.griffinlim(
            specgram,
            self.window,
            self.n_fft,
            self.hop_length,
            self.win_length,
            self.power,
            self.n_iter,
            self.momentum,
            self.length,
            self.rand_init,
        )



"""shufflenet in pytorch



[1] Xiangyu Zhang, Xinyu Zhou, Mengxiao Lin, Jian Sun.

    ShuffleNet: An Extremely Efficient Convolutional Neural Network for Mobile Devices
    https://arxiv.org/abs/1707.01083v2
"""

from functools import partial

import torch
import torch.nn as nn


class BasicConv2d(nn.Module):

    def __init__(self, input_channels, output_channels, kernel_size, **kwargs):
        super().__init__()
        self.conv = nn.Conv2d(input_channels, output_channels, kernel_size, **kwargs)
        self.bn = nn.BatchNorm2d(output_channels)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x

class ChannelShuffle(nn.Module):

    def __init__(self, groups):
        super().__init__()
        self.groups = groups

    def forward(self, x):
        batchsize, channels, height, width = x.data.size()
        channels_per_group = int(channels / self.groups)

        #"""suppose a convolutional layer with g groups whose output has
        #g x n channels; we first reshape the output channel dimension
        #into (g, n)"""
        x = x.view(batchsize, self.groups, channels_per_group, height, width)

        #"""transposing and then flattening it back as the input of next layer."""
        x = x.transpose(1, 2).contiguous()
        x = x.view(batchsize, -1, height, width)

        return x

class DepthwiseConv2d(nn.Module):

    def __init__(self, input_channels, output_channels, kernel_size, **kwargs):
        super().__init__()
        self.depthwise = nn.Sequential(
            nn.Conv2d(input_channels, output_channels, kernel_size, **kwargs),
            nn.BatchNorm2d(output_channels)
        )

    def forward(self, x):
        return self.depthwise(x)

class PointwiseConv2d(nn.Module):
    def __init__(self, input_channels, output_channels, **kwargs):
        super().__init__()
        self.pointwise = nn.Sequential(
            nn.Conv2d(input_channels, output_channels, 1, **kwargs),
            nn.BatchNorm2d(output_channels)
        )

    def forward(self, x):
        return self.pointwise(x)

class ShuffleNetUnit(nn.Module):

    def __init__(self, input_channels, output_channels, stage, stride, groups):
        super().__init__()

        #"""Similar to [9], we set the number of bottleneck channels to 1/4
        #of the output channels for each ShuffleNet unit."""
        self.bottlneck = nn.Sequential(
            PointwiseConv2d(
                input_channels,
                int(output_channels / 4),
                groups=groups
            ),
            nn.ReLU(inplace=True)
        )

        #"""Note that for Stage 2, we do not apply group convolution on the first pointwise
        #layer because the number of input channels is relatively small."""
        if stage == 2:
            self.bottlneck = nn.Sequential(
                PointwiseConv2d(
                    input_channels,
                    int(output_channels / 4),
                    groups=groups
                ),
                nn.ReLU(inplace=True)
            )

        self.channel_shuffle = ChannelShuffle(groups)

        self.depthwise = DepthwiseConv2d(
            int(output_channels / 4),
            int(output_channels / 4),
            3,
            groups=int(output_channels / 4),
            stride=stride,
            padding=1
        )

        self.expand = PointwiseConv2d(
            int(output_channels / 4),
            output_channels,
            groups=groups
        )

        self.relu = nn.ReLU(inplace=True)
        self.fusion = self._add
        self.shortcut = nn.Sequential()

        #"""As for the case where ShuffleNet is applied with stride,
        #we simply make two modifications (see Fig 2 (c)):
        #(i) add a 3 × 3 average pooling on the shortcut path;
        #(ii) replace the element-wise addition with channel concatenation,
        #which makes it easy to enlarge channel dimension with little extra
        #computation cost.
        if stride != 1 or input_channels != output_channels:
            self.shortcut = nn.AvgPool2d(3, stride=2, padding=1)

            self.expand = PointwiseConv2d(
                int(output_channels / 4),
                output_channels - input_channels,
                groups=groups
            )

            self.fusion = self._cat

    def _add(self, x, y):
        return torch.add(x, y)

    def _cat(self, x, y):
        return torch.cat([x, y], dim=1)

    def forward(self, x):
        shortcut = self.shortcut(x)

        shuffled = self.bottlneck(x)
        shuffled = self.channel_shuffle(shuffled)
        shuffled = self.depthwise(shuffled)
        shuffled = self.expand(shuffled)

        output = self.fusion(shortcut, shuffled)
        output = self.relu(output)

        return output

class ShuffleNet(nn.Module):

    def __init__(self, num_blocks, num_classes=100, groups=3):
        super().__init__()

        if groups == 1:
            out_channels = [24, 144, 288, 567]
        elif groups == 2:
            out_channels = [24, 200, 400, 800]
        elif groups == 3:
            out_channels = [24, 240, 480, 960]
        elif groups == 4:
            out_channels = [24, 272, 544, 1088]
        elif groups == 8:
            out_channels = [24, 384, 768, 1536]

        self.conv1 = BasicConv2d(3, out_channels[0], 3, padding=1, stride=1)
        self.input_channels = out_channels[0]

        self.stage2 = self._make_stage(
            ShuffleNetUnit,
            num_blocks[0],
            out_channels[1],
            stride=2,
            stage=2,
            groups=groups
        )

        self.stage3 = self._make_stage(
            ShuffleNetUnit,
            num_blocks[1],
            out_channels[2],
            stride=2,
            stage=3,
            groups=groups
        )

        self.stage4 = self._make_stage(
            ShuffleNetUnit,
            num_blocks[2],
            out_channels[3],
            stride=2,
            stage=4,
            groups=groups
        )

        self.avg = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(out_channels[3], num_classes)

    def forward(self, x):
        x = self.conv1(x)
        x = self.stage2(x)
        x = self.stage3(x)
        x = self.stage4(x)
        x = self.avg(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)

        return x

    def _make_stage(self, block, num_blocks, output_channels, stride, stage, groups):
        """make shufflenet stage

        Args:
            block: block type, shuffle unit
            out_channels: output depth channel number of this stage
            num_blocks: how many blocks per stage
            stride: the stride of the first block of this stage
            stage: stage index
            groups: group number of group convolution
        Return:
            return a shuffle net stage
        """
        strides = [stride] + [1] * (num_blocks - 1)

        stage = []

        for stride in strides:
            stage.append(
                block(
                    self.input_channels,
                    output_channels,
                    stride=stride,
                    stage=stage,
                    groups=groups
                )
            )
            self.input_channels = output_channels

        return nn.Sequential(*stage)

def shufflenet():
    return ShuffleNet([4, 8, 4])







import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import MSELoss

from models.DeepLabV3.backbone import build_backbone
from models.DeepLabV3.backbone.mobilenet import DepthSeparableConv1d


class _ASPPModule(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, padding, dilation):
        super(_ASPPModule, self).__init__()
        self.atrous_conv = nn.Conv1d(in_channels, out_channels, kernel_size=kernel_size,
                                     stride=1, padding=padding, dilation=dilation, bias=False)
        self.bn = nn.BatchNorm1d(out_channels)
        self.relu = nn.LeakyReLU()

    def forward(self, x):
        x = self.atrous_conv(x)
        x = self.bn(x)
        return self.relu(x)


class ASPP(nn.Module):
    def __init__(self, backbone, output_stride):
        super(ASPP, self).__init__()
        if backbone == 'mobilenet':
            in_channels = 512
        else:
            in_channels = 2048
        if output_stride == 16:
            dilations = [1, 6, 12, 18]
        elif output_stride == 8:
            dilations = [1, 12, 24, 36]
        else:
            raise NotImplementedError

        self.aspp1 = _ASPPModule(in_channels, 256, 1, padding=0, dilation=dilations[0])
        self.aspp2 = _ASPPModule(in_channels, 256, 3, padding=dilations[1], dilation=dilations[1])
        self.aspp3 = _ASPPModule(in_channels, 256, 3, padding=dilations[2], dilation=dilations[2])
        self.aspp4 = _ASPPModule(in_channels, 256, 3, padding=dilations[3], dilation=dilations[3])

        self.global_avg_pool = nn.Sequential(nn.AdaptiveAvgPool1d(1),
                                             nn.Conv1d(in_channels, 256, 1, stride=1, bias=False),
                                             nn.BatchNorm1d(256),
                                             nn.LeakyReLU())
        self.final_layer = nn.Sequential(
            nn.Conv1d(256 * 5, 256, 1, bias=False),
            nn.BatchNorm1d(256),
            nn.LeakyReLU(),
            # nn.Dropout(0.5)
        )

    def forward(self, x):
        x1 = self.aspp1(x)
        x2 = self.aspp2(x)
        x3 = self.aspp3(x)
        x4 = self.aspp4(x)
        x5 = self.global_avg_pool(x)
        # print(x4.shape[2:])
        x5 = F.interpolate(x5, size=x1.shape[2:], mode='linear')
        # print(x.shape)
        x = torch.cat((x1, x2, x3, x4, x5), dim=1)
        # print(x.shape)
        x = self.final_layer(x)

        return x


class Decoder(nn.Module):
    def __init__(self, num_classes, backbone):
        super(Decoder, self).__init__()
        if backbone == 'mobilenet':
            # mobile low lever output channel
            low_level_out_channels = 32
            middle_level_out_channels = 64
        else:
            raise NotImplementedError
        # 对低维数据进行处理
        self.low_conv = nn.Sequential(
            nn.Conv1d(low_level_out_channels, 64, 3, stride=2, padding=1, bias=False),
            nn.BatchNorm1d(64),
            nn.ReLU())

        # 对middle维数据进行处理
        self.middle_conv = nn.Sequential(
            nn.Conv1d(middle_level_out_channels, 128, 1, bias=False),
            nn.BatchNorm1d(128),
            nn.ReLU())

        # 处理合并后的数据
        self.last_conv = nn.Sequential(
            DepthSeparableConv1d(448, 256, 3, stride=1, padding=1, bias=False),
            nn.Dropout(0.5),
            DepthSeparableConv1d(256, 256, 3, stride=1, padding=1, bias=False),
            nn.Dropout(0.2),
            nn.ConvTranspose1d(256, 256, 3, stride=2, padding=1, output_padding=1),
            nn.BatchNorm1d(256),
            nn.LeakyReLU(),
            nn.Conv1d(256, num_classes, kernel_size=1, stride=1),
            nn.Sigmoid())

    def forward(self, low_level_feat, middle_level_feat, x):
        print("final1", x.shape)
        low_level_feat = self.low_conv(low_level_feat)
        print("low", low_level_feat.shape)
        print("middle", middle_level_feat.shape)
        middle_level_feat = self.middle_conv(middle_level_feat)

        # conv after merge
        x = F.interpolate(x, size=middle_level_feat.size()[2:], mode='linear', align_corners=False)
        x = torch.cat((x, low_level_feat, middle_level_feat), dim=1)
        print("final1", x.shape)
        x = self.last_conv(x)
        print("final", x.shape)
        return x


class DeepLab(nn.Module):
    def __init__(self, backbone='mobilenet', output_stride=16, num_classes=1):
        super(DeepLab, self).__init__()
        if backbone == 'drn':
            output_stride = 8

        self.backbone = build_backbone(backbone, in_channels=1)
        self.aspp = ASPP(backbone, output_stride)
        self.decoder = Decoder(num_classes, backbone)

    def forward(self, input):
        low_level_feat, middle_level_feat, x = self.backbone(input)

        x = self.aspp(x)
        x = self.decoder(low_level_feat, middle_level_feat, x)
        # 变为与输入 size 一样
        # print(x.size())
        x = F.interpolate(x, size=input.size()[2:], mode='linear')

        return [x]

    def loss_function(self, *args, gt) -> dict:
        recons = args[0]
        recons_loss_ = MSELoss()
        recons_loss = recons_loss_(recons.flatten(), gt.flatten())
        loss = recons_loss
        return loss


if __name__ == "__main__":
    model = DeepLab(backbone='mobilenet', output_stride=16)
    model.eval()
    input = torch.rand(10, 1, 640)
    output = model(input)
    print(output[0].size())



from torchsense.models.lit_model.lit_classify import LitClassModel
from torchsense.models.lit_model.lit_regression import LitRegressModel
from torchsense.models.lit_model.lit_multimodal import LitMultimodalModel
from torchsense.models.lit_model.lit_twostage import LitTwoStageModel


import lightning as L
import torch
from torchmetrics.functional.classification.accuracy import accuracy
import torch.optim as optim
import torch.nn.functional as F
import torch.nn as nn


class LitClassModel(L.LightningModule):
    def __init__(self, model, lr=1, gamma=0.7) -> None:
        super().__init__()
        self.save_hyperparameters(ignore=['model'])
        self.model = model
        self.model.apply(self.weights_init)
        self.loss_fn = torch.nn.CrossEntropyLoss()

    def forward(self, x: torch.Tensor):
        return self.model(x)

    def _calculate_loss(self, batch, mode="train"):
        x, y = batch
        preds = self.model(x)
        # print(preds.max())
        loss = self.loss_fn(preds, y)
        acc = (preds.argmax(dim=-1) == y).float().mean()

        self.log("%s_loss" % mode, loss, prog_bar=True)
        self.log("%s_acc" % mode, acc, prog_bar=True)
        return loss

    def configure_optimizers(self):
        # optimizer = optim.AdamW(self.parameters(), lr=self.hparams.lr)
        # lr_scheduler = optim.lr_scheduler.MultiStepLR(optimizer, milestones=[100, 150], gamma=0.1)
        optimizer = optim.Adadelta(self.parameters(), lr=self.hparams.lr)
        lr_scheduler = optim.lr_scheduler.StepLR(optimizer, step_size=1, gamma=self.hparams.gamma)
        return [optimizer], [lr_scheduler]

    def training_step(self, batch, batch_idx):
        loss = self._calculate_loss(batch, mode="train")
        return loss

    def validation_step(self, batch, batch_idx):
        self._calculate_loss(batch, mode="val")

    def test_step(self, batch, batch_idx):
        self._calculate_loss(batch, mode="test")

    def weights_init(self, m):
        if isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d):
            nn.init.normal_(m.weight.data, 0.0, 0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias.data, 0)
        elif isinstance(m, nn.BatchNorm2d):
            nn.init.normal_(m.weight.data, 1.0, 0.02)
            nn.init.constant_(m.bias.data, 0)
        # 对于任何其他类型的模块，如果它有子模块，则递归地应用 weights_init 函数
        elif isinstance(m, nn.Module):
            for name, child in m.named_children():
                self.weights_init(child)



"""resnet in pytorch



[1] Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun.

    Deep Residual Learning for Image Recognition
    https://arxiv.org/abs/1512.03385v1
"""

import torch
import torch.nn as nn
from torch.autograd import Variable
# from thop import profile
# from thop import clever_format

class BasicBlock(nn.Module):
    """Basic Block for resnet 18 and resnet 34

    """

    # BasicBlock and BottleNeck block
    # have different output size
    # we use class attribute expansion
    # to distinct
    expansion = 1

    def __init__(self, in_channels, out_channels, stride=1):
        super().__init__()

        # residual function
        self.residual_function = nn.Sequential(
            nn.Conv1d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=False),
            nn.BatchNorm1d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv1d(out_channels, out_channels * BasicBlock.expansion, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm1d(out_channels * BasicBlock.expansion)
        )

        # shortcut
        self.shortcut = nn.Sequential()

        # the shortcut output dimension is not the same with residual function
        # use 1*1 convolution to match the dimension
        if stride != 1 or in_channels != BasicBlock.expansion * out_channels:
            self.shortcut = nn.Sequential(
                nn.Conv1d(in_channels, out_channels * BasicBlock.expansion, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm1d(out_channels * BasicBlock.expansion)
            )

    def forward(self, x):
        return nn.ReLU(inplace=True)(self.residual_function(x) + self.shortcut(x))


class BottleNeck(nn.Module):
    """Residual block for resnet over 50 layers

    """
    expansion = 4

    def __init__(self, in_channels, out_channels, stride=1):
        super().__init__()
        self.residual_function = nn.Sequential(
            nn.Conv1d(in_channels, out_channels, kernel_size=1, bias=False),
            nn.BatchNorm1d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv1d(out_channels, out_channels, stride=stride, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm1d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv1d(out_channels, out_channels * BottleNeck.expansion, kernel_size=1, bias=False),
            nn.BatchNorm1d(out_channels * BottleNeck.expansion),
        )

        self.shortcut = nn.Sequential()

        if stride != 1 or in_channels != out_channels * BottleNeck.expansion:
            self.shortcut = nn.Sequential(
                nn.Conv1d(in_channels, out_channels * BottleNeck.expansion, stride=stride, kernel_size=1, bias=False),
                nn.BatchNorm1d(out_channels * BottleNeck.expansion)
            )

    def forward(self, x):
        return nn.ReLU(inplace=True)(self.residual_function(x) + self.shortcut(x))


class ResNet(nn.Module):

    def __init__(self, block, num_block, num_classes=1):
        super().__init__()

        self.in_channels = 64

        self.conv1 = nn.Sequential(
            nn.Conv1d(1, 64, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm1d(64),
            nn.ReLU(inplace=True))
        # we use a different inputsize than the original paper
        # so conv2_x's stride is 1
        self.conv2_x = self._make_layer(block, 64, num_block[0], 1)
        self.conv3_x = self._make_layer(block, 128, num_block[1], 2)
        self.conv4_x = self._make_layer(block, 256, num_block[2], 2)
        self.conv5_x = self._make_layer(block, 512, num_block[3], 2)
        self.avg_pool = nn.AdaptiveAvgPool1d(1)
        self.fc = nn.Linear(512 * block.expansion, num_classes)

    def _make_layer(self, block, out_channels, num_blocks, stride):
        """make resnet layers(by layer i didnt mean this 'layer' was the
        same as a neuron netowork layer, ex. conv layer), one layer may
        contain more than one residual block

        Args:
            block: block type, basic block or bottle neck block
            out_channels: output depth channel number of this layer
            num_blocks: how many blocks per layer
            stride: the stride of the first block of this layer

        Return:
            return a resnet layer
        """

        # we have num_block blocks per layer, the first block
        # could be 1 or 2, other blocks would always be 1
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_channels, out_channels, stride))
            self.in_channels = out_channels * block.expansion

        return nn.Sequential(*layers)

    def forward(self, x):
        b_size = x.shape[0]
        output = self.conv1(x)
        output = self.conv2_x(output)
        output = self.conv3_x(output)
        output = self.conv4_x(output)
        output = self.conv5_x(output)
        output = self.avg_pool(output)
        output = output.view(output.size(0), -1)
        output = self.fc(output)
        output = output.reshape(b_size, 1, -1)

        return output

    def get_reconstruction_loss(self, *output, labels):
        """Given a batch of images, this function returns the reconstruction loss (MSE in our case)"""
        x_hat = output[0]
        # print(x_hat.shape())
        _loss = nn.SmoothL1Loss()
        loss = _loss(x_hat, labels)
        return loss


def resnet18(num_classes=1):
    """ return a ResNet 18 object
    """
    return ResNet(BasicBlock, [2, 2, 2, 2],num_classes=num_classes)


def resnet34():
    """ return a ResNet 34 object
    """
    return ResNet(BasicBlock, [3, 4, 6, 3])


def resnet50():
    """ return a ResNet 50 object
    """
    return ResNet(BottleNeck, [3, 4, 6, 3])


def resnet101():
    """ return a ResNet 101 object
    """
    return ResNet(BottleNeck, [3, 4, 23, 3])


def resnet152():
    """ return a ResNet 152 object
    """
    return ResNet(BottleNeck, [3, 8, 36, 3])


if __name__ == '__main__':
    # gradient check
    batch_size = 1
    model = resnet50().cuda()
    loss_fn = torch.nn.MSELoss()
    input = Variable(torch.randn(batch_size, 1, 3520)).cuda()
    target = Variable(torch.randn(batch_size, 1, 2)).double().cuda()
    flops, params = profile(model, inputs=(input, ))
    flops, params = clever_format([flops, params], "%.3f")
    output = model(input)
    output = output[0].double()
    res = torch.autograd.gradcheck(loss_fn, (output.flatten(), target.flatten()), eps=1e-6, raise_exception=True)
    print(flops, params)
    print(output.size())



import torch
from typing import Any, Callable, cast, Dict, List, Optional, Tuple, Union
import numpy as np
from h5py import File


def has_batch_dimension(tensor) -> bool:
    if isinstance(tensor, list) or isinstance(tensor, tuple):
        # Check if each element in the list has a valid length (3 dimensions)
        for element in tensor:
            if not hasattr(element, 'ndim') or element.ndim != 3:
                return False
        return True
    if tensor.ndim == 3:
        return True
    return False


def add_batch_dimension(tensor: Union[torch.Tensor, List[torch.Tensor], Tuple[torch.Tensor, ...]]) -> Union[
    torch.Tensor, List[torch.Tensor], Tuple[torch.Tensor, ...]]:
    # Function to unsqueeze tensor along the specified dimension
    def unsqueeze_element(element: torch.Tensor) -> torch.Tensor:
        if isinstance(element, torch.Tensor):
            return element.unsqueeze(dim=0)
        raise ValueError("All elements must be torch tensors")

    # Check if the input is a list
    if isinstance(tensor, list):
        return [unsqueeze_element(element) for element in tensor]

    # Check if the input is a tuple
    if isinstance(tensor, tuple):
        return tuple(unsqueeze_element(element) for element in tensor)

    # Check if the input is a tensor
    if isinstance(tensor, torch.Tensor):
        return tensor.unsqueeze(dim=0)

    # If the input is neither a list, tuple, nor tensor, raise an error
    raise ValueError("Input must be a torch tensor, a list of tensors, or a tuple of tensors")


def remove_batch_dimension(tensor: Union[torch.Tensor, List[torch.Tensor], Tuple[torch.Tensor, ...]]) -> Union[
    torch.Tensor, List[torch.Tensor], Tuple[torch.Tensor, ...]]:
    # Function to squeeze tensor along the specified dimension
    def squeeze_element(element: torch.Tensor) -> torch.Tensor:
        if isinstance(element, torch.Tensor):
            return element.squeeze(dim=0)
        raise ValueError("All elements must be torch tensors")

    # Check if the input is a list
    if isinstance(tensor, list):
        return [squeeze_element(element) for element in tensor]

    # Check if the input is a tuple
    if isinstance(tensor, tuple):
        return tuple(squeeze_element(element) for element in tensor)

    # Check if the input is a tensor
    if isinstance(tensor, torch.Tensor):
        return tensor.squeeze(dim=0)

    # If the input is neither a list, tuple, nor tensor, raise an error
    raise ValueError("Input must be a torch tensor, a list of tensors, or a tuple of tensors")


def load_file(file_path, keys):
    file_extension = file_path.split('.')[-1].lower()

    if file_extension == '.txt':
        return load_text_file(file_path)
    elif file_extension == 'csv':
        return load_csv_file(file_path)
    elif file_extension == 'json':
        return load_json_file(file_path)
    elif file_extension == 'mat':
        return load_mat_file(file_path, keys)
    elif file_extension == 'npz':
        return load_npz_file(file_path, keys)
    else:
        raise ValueError("Unsupported file format")


def load_text_file(file_path):
    with open(file_path, 'r') as file:
        return file.read()


def load_csv_file(file_path):
    import csv
    data = []
    with open(file_path, 'r') as file:
        csv_reader = csv.reader(file)
        for row in csv_reader:
            data.append(row)
    return data


def load_json_file(file_path):
    import json
    with open(file_path, 'r') as file:
        return json.load(file)


def load_mat_file(path, list_j):
    from scipy import io
    try:
        all_mat_data = io.loadmat(path)
        return get_meta_data(all_mat_data, list_j)

    except NotImplementedError:
        all_mat_data = File(path)
        return get_meta_data(all_mat_data, list_j)


def get_meta_data(
        raw_data: Union[Dict[str, Any], File],
        keys: Tuple[List[str], Optional[List[str]]]
) -> Any:
    def parse_key(key: str) -> Tuple[str, Optional[int]]:
        if '[' in key and ']' in key:
            key_name = key[:key.index('[')]
            row_str = key[key.index('[') + 1:key.index(']')]
            if row_str.isdigit():
                return key_name, int(row_str)
        return key, None

    if keys is None:
        # If keys is None, return all values in raw_data as a tuple
        # Remove meta info
        if '__header__' in raw_data:
            del raw_data['__header__']
        if '__version__' in raw_data:
            del raw_data['__version__']
        if '__globals__' in raw_data:
            del raw_data['__globals__']
        return tuple(raw_data.values())
    else:
        inputs = keys
        input_values = []

        # 提取inputs并处理特定行
        for key in inputs:
            key_name, row = parse_key(key)
            if key_name in raw_data:
                value = raw_data[key_name]
                if row is not None and len(value) > row:
                    if value.shape[0] > value.shape[1]:
                        input_values.append(value[:, row])
                    else:
                        input_values.append(value[row:, ])
                else:
                    input_values.append(value)

        # 将列表转换为元组
        return input_values


def load_npz_file(path, keys):
    npz_data = np.load(path)
    return tuple(npz_data[key] for key in keys if key in npz_data)

# 使用示例
# file_path = 'example.txt'
# data1 = load_file(file_path)
# print(data1)



import torch
from torch import nn
from torch.autograd import Variable

class Encoder(nn.Module):
    def __init__(self,length:int,):
        super(Encoder, self).__init__()

        modules = []
        hidden_dims = [32, 64, 128, 256, 512]
        map_size = int(length / pow(2, len(hidden_dims)))
        in_channels = 1
        for h_dim in hidden_dims:
            length = length / 2
            modules.append(nn.Sequential(nn.Conv1d(in_channels, out_channels=h_dim,
                                                   kernel_size=3, stride=2, padding=1),
                                         nn.LayerNorm(int(length)),
                                         nn.LeakyReLU()))
            in_channels = h_dim
        self.net = nn.Sequential(*modules)
        
        self.final_layer = nn.Sequential(nn.Conv1d(hidden_dims[-1],hidden_dims[-2],
                                               kernel_size=3,
                                               stride=2,
                                               padding=1),
                            nn.BatchNorm1d(hidden_dims[-2]),
                            nn.LeakyReLU(),
                            nn.Conv1d(hidden_dims[-2], out_channels= 1,
                                      kernel_size= 3, padding= 1),
                            nn.Dropout(0.2),
                            nn.Linear(40,2),
                           )
    def forward(self, x):
        x = self.net(x)
        # x = self.fc(x)
        x =self.final_layer(x)
        return x
class Encoder2D(nn.Module):
    def __init__(self,height:int,width:int):
        super(Encoder2D, self).__init__()

        modules = []
        hidden_dims = [32, 64, 128, 256, 512]
        
        in_channels = 1
        for h_dim in hidden_dims:
            height = int(height / 2)
            width = int(width /2)
            modules.append(nn.Sequential(nn.Conv2d(in_channels, out_channels=h_dim,
                                                   kernel_size=3, stride=2, padding=1),
                                         nn.LayerNorm([int(height), int(width)]),
                                         nn.LeakyReLU()))
            in_channels = h_dim
        self.net = nn.Sequential(*modules)
        
        self.final_layer = nn.Sequential(nn.Conv2d(hidden_dims[-1],hidden_dims[-2],
                                               kernel_size=3,
                                               stride=2,
                                               padding=1),
                            nn.BatchNorm2d(hidden_dims[-2]),
                            nn.LeakyReLU(),
                            nn.Conv2d(hidden_dims[-2], out_channels= 1,
                                      kernel_size= 3, padding= 1),
                            # nn.Dropout(0.2),
                            # nn.Linear(40,2),
                           )
    def forward(self, x):
        x = self.net(x)
        # x = self.fc(x)
        x =self.final_layer(x)
        return x
class lstm(nn.Module):
    def __init__(self):
        super(lstm, self).__init__()
        self.hidden_size = 128
        self.time_step = 32
        self.length = 3520
        self.input_size = int(self.length / self.time_step)
        # 输入的维度（input size）指的是每个时间步输入的特征向量的维度
        self.rnn = nn.LSTM(
            input_size=self.input_size,
            hidden_size=self.hidden_size,
            num_layers=2,
            batch_first=True,
        )
        self.encoder = Encoder(self.hidden_size*self.time_step)
        self.encoder2d = Encoder2D(self.time_step,self.hidden_size)
    def forward(self, x):
        # print('0',x.shape,len(x))
        # (batch_size, sequence_length, input_size) 这里序列长度含义和时间步含义重合
        x = x.view(-1, self.time_step, self.input_size)
        # print('1',x.shape )
        r_out, (h_n, h_c) = self.rnn(x, None)
        # batch_size ,time step,hidden_size
        b, l, h = r_out.shape
        # print(r_out.shape )
        r_out = r_out.reshape(b,-1,l,h)
        # print(r_out.shape )
        out = self.encoder2d(r_out)
        # print(out.shape)
        return out


class gru(nn.Module):
    def __init__(self):
        super(gru, self).__init__()
        self.hidden_size = 128
        self.time_step = 10
        self.rnn = nn.GRU(
            input_size=352,
            hidden_size=self.hidden_size,
            num_layers=2,
            batch_first=True,
        )

        self.out = nn.Sequential(nn.Tanh(),
                                 nn.Linear(self.hidden_size, 1024),
                                 nn.Dropout(0.5),
                                 nn.Linear(1024, 2))  # 最后时刻的hidden映射

    def forward(self, x):
        x = x.view(len(x), self.time_step, -1)

        r_out, h_n = self.rnn(x, None)
        b, l, h = r_out.shape
        # print(r_out.shape )
        r_out = r_out.reshape(l * b, h)
        # print(r_out.shape )
        out = self.out(r_out)
        # print('2',b,out.shape )
        out = out.unsqueeze(1)
        out = out.reshape(b, 1, -1)
        return out


if __name__ == "__main__":
    # gradient check
    batch_size = 32
    model = lstm().cuda()
    loss_fn = torch.nn.MSELoss()
    input = Variable(torch.randn(batch_size, 1, 3520)).cuda()
    target = Variable(torch.randn(batch_size, 1, 3520)).cuda()

    output = model(input)
    print(output.shape)
    # output = output[0][0].double()
    # res = torch.autograd.gradcheck(loss_fn, (output, target.squeeze()), eps=1e-6, raise_exception=True)
    print(model)
    print(output.size())



"""squeezenet in pytorch



[1] Song Han, Jeff Pool, John Tran, William J. Dally

    squeezenet: Learning both Weights and Connections for Efficient Neural Networks
    https://arxiv.org/abs/1506.02626
"""

import torch
import torch.nn as nn


class Fire(nn.Module):

    def __init__(self, in_channel, out_channel, squzee_channel):

        super().__init__()
        self.squeeze = nn.Sequential(
            nn.Conv1d(in_channel, squzee_channel, 1),
            nn.BatchNorm1d(squzee_channel),
            nn.ReLU(inplace=True)
        )

        self.expand_1x1 = nn.Sequential(
            nn.Conv1d(squzee_channel, int(out_channel / 2), 1),
            nn.BatchNorm1d(int(out_channel / 2)),
            nn.ReLU(inplace=True)
        )

        self.expand_3x3 = nn.Sequential(
            nn.Conv1d(squzee_channel, int(out_channel / 2), 3, padding=1),
            nn.BatchNorm1d(int(out_channel / 2)),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):

        x = self.squeeze(x)
        x = torch.cat([
            self.expand_1x1(x),
            self.expand_3x3(x)
        ], 1)

        return x

class SqueezeNet(nn.Module):

    """mobile net with simple bypass"""
    def __init__(self, class_num=2):

        super().__init__()
        self.stem = nn.Sequential(
            nn.Conv1d(1, 96, 3, padding=1),
            nn.BatchNorm1d(96),
            nn.ReLU(inplace=True),
            nn.MaxPool1d(2, 2)
        )

        self.fire2 = Fire(96, 128, 16)
        self.fire3 = Fire(128, 128, 16)
        self.fire4 = Fire(128, 256, 32)
        self.fire5 = Fire(256, 256, 32)
        self.fire6 = Fire(256, 384, 48)
        self.fire7 = Fire(384, 384, 48)
        self.fire8 = Fire(384, 512, 64)
        self.fire9 = Fire(512, 512, 64)

        self.conv10 = nn.Conv1d(512, class_num, 1)
        self.avg = nn.AdaptiveAvgPool1d(1)
        self.maxpool = nn.MaxPool1d(2, 2)

    def forward(self, x):
        x = self.stem(x)

        f2 = self.fire2(x)
        f3 = self.fire3(f2) + f2
        f4 = self.fire4(f3)
        f4 = self.maxpool(f4)

        f5 = self.fire5(f4) + f4
        f6 = self.fire6(f5)
        f7 = self.fire7(f6) + f6
        f8 = self.fire8(f7)
        f8 = self.maxpool(f8)

        f9 = self.fire9(f8)
        c10 = self.conv10(f9)

        x = self.avg(c10)
        x = x.view(x.size(0), -1)

        return x

def squeezenet(class_num=2):
    return SqueezeNet(class_num=class_num)




import time

import torch
import torch.nn as nn
from einops import rearrange


class BasicBlock(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, padding=None):
        super(BasicBlock, self).__init__()
        self.basic = nn.Sequential(
            nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, padding=padding),
            nn.BatchNorm2d(out_channels),
            nn.LeakyReLU(inplace=True)
        )

    def forward(self, x):
        return self.basic(x)


class DSConv1d(nn.Module):

    def __init__(self, input_channels, out_channels, kernel_size, **kwargs):
        super().__init__()
        self.Depthwise = nn.Sequential(
            nn.Conv1d(input_channels, input_channels,
                      kernel_size, groups=input_channels, **kwargs),
            nn.BatchNorm1d(input_channels),
            nn.LeakyReLU(inplace=True)
        )

        self.Pointwise = nn.Sequential(
            nn.Conv1d(input_channels, out_channels, 1),
            nn.BatchNorm1d(out_channels),
            nn.LeakyReLU(inplace=True)
        )

    def forward(self, x):
        x = self.Depthwise(x)
        x = self.Pointwise(x)
        return x


class DSConv2d(nn.Module):

    def __init__(self, input_channels, out_channels, kernel_size, **kwargs):
        super().__init__()
        self.Depthwise = nn.Sequential(
            nn.Conv2d(input_channels, input_channels,
                      kernel_size, groups=input_channels, **kwargs),
            nn.BatchNorm2d(input_channels),
            nn.LeakyReLU(inplace=True)
        )

        self.Pointwise = nn.Sequential(
            nn.Conv2d(input_channels, out_channels, 1),
            nn.BatchNorm2d(out_channels),
            nn.LeakyReLU(inplace=True)
        )

    def forward(self, x):
        x = self.Depthwise(x)
        x = self.Pointwise(x)
        return x


class LinearBottleNeck(nn.Module):

    def __init__(self, in_channels, out_channels, stride, t=6, class_num=100):
        super().__init__()

        self.residual = nn.Sequential(
            # pw
            nn.Conv1d(in_channels, in_channels * t, 1, bias=False),
            nn.BatchNorm1d(in_channels * t),
            nn.ReLU6(inplace=True),
            # dw
            nn.Conv1d(in_channels * t, in_channels * t,
                      3, stride=stride, padding=1, groups=in_channels * t, bias=False),
            nn.BatchNorm1d(in_channels * t),
            nn.ReLU6(inplace=True),
            # pw
            nn.Conv1d(in_channels * t, out_channels, 1, bias=False),
            nn.BatchNorm1d(out_channels)
        )

        self.stride = stride
        self.in_channels = in_channels
        self.out_channels = out_channels

    def forward(self, x):
        residual = self.residual(x)
        if self.stride == 1 and self.in_channels == self.out_channels:
            residual += x

        return residual


class FTB(nn.Module):

    def __init__(self, input_dim=257, in_channel=96, r_channel=5):
        super(FTB, self).__init__()

        self.in_channel = in_channel
        self.conv1 = nn.Sequential(
            nn.Conv2d(in_channel, r_channel, kernel_size=(1, 1)),
            nn.BatchNorm2d(r_channel),
            nn.ReLU()
        )

        self.conv1d = nn.Sequential(
            nn.Conv1d(r_channel * input_dim, in_channel, kernel_size=9, padding=4),
            nn.BatchNorm1d(in_channel),
            nn.ReLU()
        )

        self.freq_fc = nn.Linear(input_dim, input_dim, bias=False)
        self.conv2 = nn.Sequential(
            nn.Conv2d(in_channel * 2, in_channel, kernel_size=(1, 1)),
            nn.BatchNorm2d(in_channel),
            nn.ReLU()
        )

    def forward(self, inputs):
        '''
        inputs should be [Batch, Ca, Dim, Time]
        '''

        # T-F attention
        conv1_out = self.conv1(inputs)

        B, C, D, T = conv1_out.size()
        reshape1_out = torch.reshape(conv1_out, [B, C * D, T])
        conv1d_out = self.conv1d(reshape1_out)
        conv1d_out = torch.reshape(conv1d_out, [B, self.in_channel, 1, T])

        # now is also [B,C,D,T]
        att_out = conv1d_out * inputs

        # tranpose to [B,C,T,D]
        att_out = torch.transpose(att_out, 2, 3)
        freqfc_out = self.freq_fc(att_out)
        att_out = torch.transpose(freqfc_out, 2, 3)

        cat_out = torch.cat([att_out, inputs], 1)
        outputs = self.conv2(cat_out)
        return outputs


class TFS_AttConv(nn.Module):
    def __init__(self, input_dim=257, channel_amp=96):
        super().__init__()
        self.ftb1 = FTB(input_dim=input_dim,
                        in_channel=channel_amp,
                        )
        self.amp_conv1 = nn.Sequential(
            nn.Conv2d(channel_amp, channel_amp, kernel_size=5, padding=2),
            nn.BatchNorm2d(channel_amp),
            nn.LeakyReLU()
        )

        self.amp_conv2 = nn.Sequential(
            nn.Conv2d(channel_amp, channel_amp, kernel_size=(1, 25), padding=(0, 12)),
            nn.BatchNorm2d(channel_amp),
            nn.LeakyReLU()
        )

        self.amp_conv3 = nn.Sequential(
            nn.Conv2d(channel_amp, channel_amp, kernel_size=5, padding=2),
            nn.BatchNorm2d(channel_amp),
            nn.LeakyReLU()
        )

        self.ftb2 = FTB(input_dim=input_dim, in_channel=channel_amp)

    def forward(self, amp):
        '''
        amp should be [Batch, Ca, Dim, Time]
        amp should be [Batch, Cr, Dim, Time]
        '''

        amp_out = self.ftb1(amp)
        amp_out = self.amp_conv1(amp_out)
        amp_out = self.amp_conv2(amp_out)
        amp_out = self.amp_conv3(amp_out)
        amp_out = self.ftb2(amp_out)
        return amp_out


class TF_Domain_Noise(nn.Module):
    def __init__(self, channel_amp=96):
        super(TF_Domain_Noise, self).__init__()
        self.amp_conv1 = nn.Sequential(
            nn.Conv2d(1, channel_amp,
                      kernel_size=[1, 7],
                      padding=(0, 3)
                      ),
            nn.BatchNorm2d(channel_amp),
            nn.LeakyReLU(),
            nn.Conv2d(channel_amp, channel_amp,
                      kernel_size=[7, 1],
                      padding=(3, 0)
                      ),
            nn.BatchNorm2d(channel_amp),
            nn.LeakyReLU(),
        )
        self.amp_conv2 = nn.Sequential(
            nn.Conv2d(channel_amp, 1, kernel_size=[1, 1]),
            nn.BatchNorm2d(1),
            nn.LeakyReLU(),
        )
        self.tfss = nn.ModuleList()
        for idx in range(3):
            self.tfss.append(TFS_AttConv())

    def forward(self, amp):
        time_start = time.time()
        spec = self.amp_conv1(amp)

        s_spec = [spec]
        for idx, layer in enumerate(self.tfss):
            if idx != 0:
                s_spec.append(s_spec[0] + s_spec[-1])
            s_spec.append(layer(s_spec[-1]))
        out = self.amp_conv2(s_spec[-1])
        # s_spec = spec
        # for idx, layer in enumerate(self.tfss):
        #     if idx != 0:
        #         spec += s_spec
        #     spec = layer(spec)
        # out = self.amp_conv2(spec)

        return out


class TFU_Conv(nn.Module):
    def __init__(self, in_channels=96):
        super(TFU_Conv, self).__init__()
        self.module = nn.Sequential(
            BasicBlock(in_channels=in_channels, kernel_size=(3, 5), out_channels=in_channels, padding=(1, 2)),
            BasicBlock(in_channels=in_channels, kernel_size=(9, 1), out_channels=in_channels, padding=(4, 0)),
            BasicBlock(in_channels=in_channels, kernel_size=(3, 5), out_channels=in_channels, padding=(1, 2), )
        )

    def forward(self, x):
        return self.module(x)


class TF_Domain_ACC(nn.Module):
    def __init__(self, channel_amp=96):
        super(TF_Domain_ACC, self).__init__()
        self.conv1 = nn.Sequential(
            nn.Conv2d(1, channel_amp,
                      kernel_size=(1, 5),
                      padding=(0, 2)
                      ),
            nn.BatchNorm2d(channel_amp),
            nn.LeakyReLU(),
            nn.Conv2d(channel_amp, channel_amp,
                      kernel_size=(7, 1),
                      padding=(3, 0)
                      ),
            nn.BatchNorm2d(channel_amp),
            nn.LeakyReLU(),
        )
        self.tfus = nn.ModuleList()
        for idx in range(3):
            self.tfus.append(TFU_Conv())
        self.conv2 = nn.Sequential(
            nn.Conv2d(channel_amp, 1, kernel_size=(1, 1)),
            nn.BatchNorm2d(1),
            nn.LeakyReLU(),
        )

    def forward(self, x):
        time_start = time.time()
        spec = self.conv1(x)
        # print("generator-acc:1 conv layers"+str(time_start-time.time()))
        # time_start=time.time()
        s_spec = []
        s_spec.append(spec)

        for idx, layer in enumerate(self.tfus):
            if idx != 0:
                s_spec.append(s_spec[0] + s_spec[-1])
            s_spec.append(layer(s_spec[-1]))
        # print("generator-acc:3 TFU conv layer"+str(time_start-time.time()))
        # time_start=time.time() 
        out = self.conv2(s_spec[-1])
        # print("generator-acc:1 conv layer"+str(time_start-time.time()))
        # time_start=time.time() 
        return out


# 输入的是 带噪声的时频谱（amp）和加速计的时频谱(acc)（1，257，498）  （1，51，498）
# 输出是  增强的amp  groudtruth 纯净的amp
class Generator(nn.Module):
    def __init__(self, lstm_nums=300):
        super().__init__()
        self.noise = TF_Domain_Noise()
        self.acc = TF_Domain_ACC()
        # 注意力融合--
        self.att = FTB(input_dim=257 + 51, in_channel=1)
        # TODO 自定义实现lstm
        self.lstm = nn.LSTM(input_size=(257 + 51) * 1, hidden_size=lstm_nums, batch_first=True, bidirectional=True)
        self.fc = nn.Sequential(
            nn.Linear(lstm_nums * 2, 600),
            nn.LeakyReLU(),
            nn.Linear(600, 600),
            nn.LeakyReLU(),
            nn.Linear(600, 257),
            
        )

    def forward(self, inputs, training=True):
        acc, noise = inputs
        noise_extraction = self.noise(noise)
        acc_extraction = self.acc(acc)
        features = torch.concat((noise_extraction, acc_extraction), dim=2)
        att = self.att(features)
        att = rearrange(att, 'b 1 f t -> b t f')
        lstm = self.lstm(att)[0]
        out = self.fc(lstm)
        if not training:
            print(f"out:{out.max()}")
            print(f"noise:{noise.max()}")
        out = out.float()
        out = torch.sigmoid(out)
        mask = rearrange(out, 'b t f -> b 1 f t')
        output = mask * noise

        return output



"""shufflenetv2 in pytorch



[1] Ningning Ma, Xiangyu Zhang, Hai-Tao Zheng, Jian Sun

    ShuffleNet V2: Practical Guidelines for Efficient CNN Architecture Design
    https://arxiv.org/abs/1807.11164
"""

import torch
import torch.nn as nn
import torch.nn.functional as F


def channel_split(x, split):
    """split a tensor into two pieces along channel dimension
    Args:
        x: input tensor
        split:(int) channel size for each pieces
    """
    assert x.size(1) == split * 2
    return torch.split(x, split, dim=1)

def channel_shuffle(x, groups):
    """channel shuffle operation
    Args:
        x: input tensor
        groups: input branch number
    """

    batch_size, channels, height, width = x.size()
    channels_per_group = int(channels // groups)

    x = x.view(batch_size, groups, channels_per_group, height, width)
    x = x.transpose(1, 2).contiguous()
    x = x.view(batch_size, -1, height, width)

    return x

class ShuffleUnit(nn.Module):

    def __init__(self, in_channels, out_channels, stride):
        super().__init__()

        self.stride = stride
        self.in_channels = in_channels
        self.out_channels = out_channels

        if stride != 1 or in_channels != out_channels:
            self.residual = nn.Sequential(
                nn.Conv2d(in_channels, in_channels, 1),
                nn.BatchNorm2d(in_channels),
                nn.ReLU(inplace=True),
                nn.Conv2d(in_channels, in_channels, 3, stride=stride, padding=1, groups=in_channels),
                nn.BatchNorm2d(in_channels),
                nn.Conv2d(in_channels, int(out_channels / 2), 1),
                nn.BatchNorm2d(int(out_channels / 2)),
                nn.ReLU(inplace=True)
            )

            self.shortcut = nn.Sequential(
                nn.Conv2d(in_channels, in_channels, 3, stride=stride, padding=1, groups=in_channels),
                nn.BatchNorm2d(in_channels),
                nn.Conv2d(in_channels, int(out_channels / 2), 1),
                nn.BatchNorm2d(int(out_channels / 2)),
                nn.ReLU(inplace=True)
            )
        else:
            self.shortcut = nn.Sequential()

            in_channels = int(in_channels / 2)
            self.residual = nn.Sequential(
                nn.Conv2d(in_channels, in_channels, 1),
                nn.BatchNorm2d(in_channels),
                nn.ReLU(inplace=True),
                nn.Conv2d(in_channels, in_channels, 3, stride=stride, padding=1, groups=in_channels),
                nn.BatchNorm2d(in_channels),
                nn.Conv2d(in_channels, in_channels, 1),
                nn.BatchNorm2d(in_channels),
                nn.ReLU(inplace=True)
            )


    def forward(self, x):

        if self.stride == 1 and self.out_channels == self.in_channels:
            shortcut, residual = channel_split(x, int(self.in_channels / 2))
        else:
            shortcut = x
            residual = x

        shortcut = self.shortcut(shortcut)
        residual = self.residual(residual)
        x = torch.cat([shortcut, residual], dim=1)
        x = channel_shuffle(x, 2)

        return x

class ShuffleNetV2(nn.Module):

    def __init__(self, ratio=1, class_num=100):
        super().__init__()
        if ratio == 0.5:
            out_channels = [48, 96, 192, 1024]
        elif ratio == 1:
            out_channels = [116, 232, 464, 1024]
        elif ratio == 1.5:
            out_channels = [176, 352, 704, 1024]
        elif ratio == 2:
            out_channels = [244, 488, 976, 2048]
        else:
            ValueError('unsupported ratio number')

        self.pre = nn.Sequential(
            nn.Conv2d(3, 24, 3, padding=1),
            nn.BatchNorm2d(24)
        )

        self.stage2 = self._make_stage(24, out_channels[0], 3)
        self.stage3 = self._make_stage(out_channels[0], out_channels[1], 7)
        self.stage4 = self._make_stage(out_channels[1], out_channels[2], 3)
        self.conv5 = nn.Sequential(
            nn.Conv2d(out_channels[2], out_channels[3], 1),
            nn.BatchNorm2d(out_channels[3]),
            nn.ReLU(inplace=True)
        )

        self.fc = nn.Linear(out_channels[3], class_num)

    def forward(self, x):
        x = self.pre(x)
        x = self.stage2(x)
        x = self.stage3(x)
        x = self.stage4(x)
        x = self.conv5(x)
        x = F.adaptive_avg_pool2d(x, 1)
        x = x.view(x.size(0), -1)
        x = self.fc(x)

        return x

    def _make_stage(self, in_channels, out_channels, repeat):
        layers = []
        layers.append(ShuffleUnit(in_channels, out_channels, 2))

        while repeat:
            layers.append(ShuffleUnit(out_channels, out_channels, 1))
            repeat -= 1

        return nn.Sequential(*layers)

def shufflenetv2():
    return ShuffleNetV2()








import torch

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')


def load_from_ckpt(model, path,load_location=None):
    """
    Load model weights from a checkpoint file.

    Args:
        model (nn.Module): The model to load the weights into.
        path (str): The path to the checkpoint file.

    Returns:
        nn.Module: The model with loaded weights.
    """
    if load_location == 'cpu':
        checkpoint = torch.load(path,map_location='cpu')
    else:
        checkpoint = torch.load(path, map_location=device)
    state_dict = checkpoint["state_dict"]
    new_state_dict = {k.replace("model.", ""): v for k, v in state_dict.items()}

    # Load the new state dictionary into the model
    model.load_state_dict(new_state_dict)
    return model



from torch import nn
import torch
import numpy as np


class ToTensor:
    """Convert a PIL Image or ndarray to tensor and scale the values accordingly.

    This transform does not support torchscript.

   """

    def __call__(self, pic):
        """
        Args:
            pic (PIL Image or numpy.ndarray): Image to be converted to tensor.

        Returns:
            Tensor: Converted image.
        """
        if isinstance(pic, np.ndarray):
            # handle numpy array
            pic = torch.from_numpy(pic.reshape(1, -1).astype(np.float32))
            # backward compatibility
            return pic
        else:
            return torch.tensor(pic, dtype=torch.float32)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"



import torch.nn as nn
import torch.nn.functional as F
import torch


class CNN4(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.conv1 = nn.Conv2d(1, 32, 3, 1)
        self.conv2 = nn.Conv2d(32, 64, 3, 1)
        self.dropout1 = nn.Dropout(0.25)
        self.dropout2 = nn.Dropout(0.5)
        self.fc1 = nn.Linear(9216, 128)
        self.fc2 = nn.Linear(128, 10)

    def forward(self, x):
        x = self.conv1(x)
        x = F.relu(x)
        x = self.conv2(x)
        x = F.relu(x)
        x = F.max_pool2d(x, 2)
        x = self.dropout1(x)
        x = torch.flatten(x, 1)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.dropout2(x)
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)



"""google net in pytorch



[1] Christian Szegedy, Wei Liu, Yangqing Jia, Pierre Sermanet, Scott Reed,
    Dragomir Anguelov, Dumitru Erhan, Vincent Vanhoucke, Andrew Rabinovich.

    Going Deeper with Convolutions
    https://arxiv.org/abs/1409.4842v1
"""

import torch
import torch.nn as nn

class Inception(nn.Module):
    def __init__(self, input_channels, n1x1, n3x3_reduce, n3x3, n5x5_reduce, n5x5, pool_proj):
        super().__init__()

        #1x1conv branch
        self.b1 = nn.Sequential(
            nn.Conv2d(input_channels, n1x1, kernel_size=1),
            nn.BatchNorm2d(n1x1),
            nn.ReLU(inplace=True)
        )

        #1x1conv -> 3x3conv branch
        self.b2 = nn.Sequential(
            nn.Conv2d(input_channels, n3x3_reduce, kernel_size=1),
            nn.BatchNorm2d(n3x3_reduce),
            nn.ReLU(inplace=True),
            nn.Conv2d(n3x3_reduce, n3x3, kernel_size=3, padding=1),
            nn.BatchNorm2d(n3x3),
            nn.ReLU(inplace=True)
        )

        #1x1conv -> 5x5conv branch
        #we use 2 3x3 conv filters stacked instead
        #of 1 5x5 filters to obtain the same receptive
        #field with fewer parameters
        self.b3 = nn.Sequential(
            nn.Conv2d(input_channels, n5x5_reduce, kernel_size=1),
            nn.BatchNorm2d(n5x5_reduce),
            nn.ReLU(inplace=True),
            nn.Conv2d(n5x5_reduce, n5x5, kernel_size=3, padding=1),
            nn.BatchNorm2d(n5x5, n5x5),
            nn.ReLU(inplace=True),
            nn.Conv2d(n5x5, n5x5, kernel_size=3, padding=1),
            nn.BatchNorm2d(n5x5),
            nn.ReLU(inplace=True)
        )

        #3x3pooling -> 1x1conv
        #same conv
        self.b4 = nn.Sequential(
            nn.MaxPool2d(3, stride=1, padding=1),
            nn.Conv2d(input_channels, pool_proj, kernel_size=1),
            nn.BatchNorm2d(pool_proj),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        return torch.cat([self.b1(x), self.b2(x), self.b3(x), self.b4(x)], dim=1)


class GoogleNet(nn.Module):

    def __init__(self, num_class=100):
        super().__init__()
        self.prelayer = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
            nn.Conv2d(64, 64, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
            nn.Conv2d(64, 192, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(192),
            nn.ReLU(inplace=True),
        )

        #although we only use 1 conv layer as prelayer,
        #we still use name a3, b3.......
        self.a3 = Inception(192, 64, 96, 128, 16, 32, 32)
        self.b3 = Inception(256, 128, 128, 192, 32, 96, 64)

        ##"""In general, an Inception network is a network consisting of
        ##modules of the above type stacked upon each other, with occasional
        ##max-pooling layers with stride 2 to halve the resolution of the
        ##grid"""
        self.maxpool = nn.MaxPool2d(3, stride=2, padding=1)

        self.a4 = Inception(480, 192, 96, 208, 16, 48, 64)
        self.b4 = Inception(512, 160, 112, 224, 24, 64, 64)
        self.c4 = Inception(512, 128, 128, 256, 24, 64, 64)
        self.d4 = Inception(512, 112, 144, 288, 32, 64, 64)
        self.e4 = Inception(528, 256, 160, 320, 32, 128, 128)

        self.a5 = Inception(832, 256, 160, 320, 32, 128, 128)
        self.b5 = Inception(832, 384, 192, 384, 48, 128, 128)

        #input feature size: 8*8*1024
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.dropout = nn.Dropout2d(p=0.4)
        self.linear = nn.Linear(1024, num_class)

    def forward(self, x):
        x = self.prelayer(x)
        x = self.maxpool(x)
        x = self.a3(x)
        x = self.b3(x)

        x = self.maxpool(x)

        x = self.a4(x)
        x = self.b4(x)
        x = self.c4(x)
        x = self.d4(x)
        x = self.e4(x)

        x = self.maxpool(x)

        x = self.a5(x)
        x = self.b5(x)

        #"""It was found that a move from fully connected layers to
        #average pooling improved the top-1 accuracy by about 0.6%,
        #however the use of dropout remained essential even after
        #removing the fully connected layers."""
        x = self.avgpool(x)
        x = self.dropout(x)
        x = x.view(x.size()[0], -1)
        x = self.linear(x)

        return x

def googlenet():
    return GoogleNet()





# test_unet.py

import torch
import pytest
from torchsense.models import UNet, resnet18

n_channels = 1


def test_unet_forward():
    # 定义模型参数

    n_classes = 2
    bilinear = True

    batch_size = 4
    seq_length = 128
    # 创建模型实例
    model = UNet(n_channels=n_channels, n_classes=n_classes, bilinear=bilinear)
    # 创建一个随机的输入张量，形状为 (batch_size, n_channels, seq_length)
    input_tensor = torch.randn(batch_size, n_channels, seq_length)

    # 运行模型前向传播
    output_tensor = model(input_tensor)

    # 检查输出张量的形状是否正确
    assert output_tensor.shape == (batch_size, n_classes, seq_length), \
        f"Expected output shape (batch_size, n_classes, seq_length), but got {output_tensor.shape}"


def test_resnet_forward():
    batch_size = 4

    seq_length = 128

    # 创建一个随机的输入张量，形状为 (batch_size, n_channels, seq_length)
    input_tensor = torch.randn(batch_size, n_channels, seq_length)
    model = resnet18(128)
    # 运行模型前向传播
    output_tensor = model(input_tensor)

    # 检查输出张量的形状是否正确
    assert output_tensor.shape == (batch_size, n_channels, seq_length), \
        f"Expected output shape (batch_size, n_classes, seq_length), but got {output_tensor.shape}"


# 运行测试
if __name__ == "__main__":
    pytest.main()



import numpy as np
import torch
import numpy.random as random
from .utils import add_batch_dimension,has_batch_dimension,remove_batch_dimension

class PreModule(torch.nn.Module):
    """
    A module that applies preprocessing operations to input data before passing it through a model.
    
    Args:
        model (torch.nn.Module): The model to be applied to the input data.
    """
    def __init__(self, model):
        super().__init__()
        self.model = model
        self.model.eval()
        
    def forward(self, x):
        """
        Forward pass of the PreModule.
        
        Args:
            x (torch.Tensor): The input data.
        
        Returns:
            torch.Tensor: The output of the model after applying preprocessing operations.
        """
        if not has_batch_dimension(x):
            x = add_batch_dimension(x)
        x = self.model(x)
        x = x.detach()
        if has_batch_dimension:
            remove_batch_dimension(x)
        return x
